import numpy as np

from pymoo.util.misc import at_least_2d_array, to_1d_array_if_possible


class Decomposition:

    def __init__(self, eps=0.0, _type="auto", **kwargs) -> None:
        super().__init__()
        self.eps = eps
        self._type = _type
        self.ideal_point, self.utopian_point, self.nadir_point = None, None, None

    def __call__(self, *args, **kwargs):
        return self.do(*args, **kwargs)

    def do(self,
           F,
           weights,
           _type="auto",
           ideal_point=None,
           utopian_point=None,
           nadir_point=None,
           **kwargs):

        _F, _weights = to_1d_array_if_possible(F), to_1d_array_if_possible(weights)

        if _type == "auto":
            if _F.ndim == 1 and _weights.ndim > 1:
                _type = "one_to_many"
            elif _F.ndim > 1 and _weights.ndim == 1:
                _type = "many_to_one"
            elif _F.ndim == 2 and _weights.ndim == 2 and _F.shape[0] == _weights.shape[0]:
                _type = "one_to_one"
            else:
                _type = "many_to_many"

        # make both at least 2d arrays
        F, weights = at_least_2d_array(F), at_least_2d_array(weights)

        # get the number of points and weights
        n_points, n_weights = F.shape[0], weights.shape[0]

        self.ideal_point = ideal_point
        if self.ideal_point is None:
            self.ideal_point = np.zeros(F.shape[1])

        self.utopian_point = utopian_point
        if self.utopian_point is None:
            self.utopian_point = self.ideal_point - self.eps

        # set the nadir point by default to value or default
        self.nadir_point = nadir_point
        if self.nadir_point is None:
            self.nadir_point = self.utopian_point + np.ones(F.shape[1])

        if _type == "one_to_one":
            D = self._do(F, weights=weights, **kwargs).flatten()

        elif _type == "one_to_many":
            F = np.repeat(F, n_weights, axis=0)
            D = self._do(F, weights=weights, **kwargs).flatten()

        elif _type == "many_to_one":
            weights = np.repeat(weights, n_points, axis=0)
            D = self._do(F, weights=weights, **kwargs).flatten()

        elif _type == "many_to_many":
            F = np.repeat(F, n_weights, axis=0)
            weights = np.tile(weights, (n_points, 1))
            D = self._do(F, weights=weights, **kwargs).reshape(n_points, n_weights)

        else:
            raise Exception("Unknown type for decomposition: %s" % _type)

        return D



import numpy as np

from pymoo.core.decision_making import DecisionMaking


class PseudoWeights(DecisionMaking):

    def __init__(self, weights, **kwargs) -> None:
        super().__init__(**kwargs)
        self.weights = weights

    def _do(self, F, return_pseudo_weights=False, **kwargs):
        ideal, nadir = self.ideal, self.nadir

        if ideal is None:
            ideal = F.min(axis=0)
        if nadir is None:
            nadir = F.max(axis=0)

        # normalized distance to the worst solution
        pseudo_weights = ((nadir - F) / (nadir - ideal))

        # normalize weights to sum up to one
        pseudo_weights = pseudo_weights / np.sum(pseudo_weights, axis=1)[:, None]

        # search for the closest individual having this pseudo weights
        I = np.argmin(np.sum(np.abs(pseudo_weights - self.weights), axis=1))

        if return_pseudo_weights:
            return I, pseudo_weights
        else:
            return I



import numpy as np

from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival
from pymoo.core.algorithm import LoopwiseAlgorithm
from pymoo.core.initialization import Initialization
from pymoo.core.population import Population
from pymoo.core.repair import NoRepair
from pymoo.core.replacement import is_better
from pymoo.core.variable import Real, Integer, get
from pymoo.docs import parse_doc_string
from pymoo.operators.crossover.pcx import PCX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.selection.rnd import fast_fill_random
from pymoo.util.display.single import SingleObjectiveOutput


# =========================================================================================================
# Implementation
# =========================================================================================================


class G3PCX(LoopwiseAlgorithm):

    def __init__(self,
                 pop_size=100,
                 sampling=FloatRandomSampling(),
                 n_offsprings=2,
                 n_parents=3,
                 family_size=2,
                 repair=NoRepair(),
                 output=SingleObjectiveOutput(),
                 **kwargs):

        super().__init__(output=output, **kwargs)

        self.pop_size = Integer(pop_size, bounds=(20, 200))
        self.repair = repair

        self.initialization = Initialization(sampling, repair=self.repair, eliminate_duplicates=False)

        self.n_offsprings = Integer(n_offsprings, bounds=(1, 10))
        self.n_parents = Integer(n_parents, bounds=(3, 10))
        self.family_size = Integer(family_size, bounds=(1, 10))

        self.crossover = PCX()
        self.crossover.prob = 1.0

        self.mutation = PM()
        self.mutation.prob = Real(0.25, bounds=(0.0, 1.0))

    def _initialize_infill(self):
        return self.initialization.do(self.problem, get(self.pop_size), algorithm=self)

    def _initialize_advance(self, infills=None, **kwargs):
        self.pop = FitnessSurvival().do(self.problem, infills, n_survive=len(infills), algorithm=self, **kwargs)

    def _next(self, **kwargs):
        pop_size, n_offsprings, n_parents = get(self.pop_size, self.n_offsprings, self.n_parents)

        # how many loops shall be iterated until one iteration has ended
        loops_per_iter = pop_size // n_offsprings

        for _ in range(loops_per_iter):

            S = np.zeros((n_offsprings, n_parents), dtype=int)
            S[:, 0] = 0
            fast_fill_random(S, len(self.pop), columns=range(1, n_parents))

            off = self.crossover(self.problem, self.pop, parents=S, algorithm=self)

            off = self.mutation(self.problem, off, algorithm=self)

            self.repair(self.problem, off, algorithm=self)

            off = yield off

            pop, family_size = self.pop, get(self.family_size)

            rnd = np.random.choice(np.arange(len(pop)), size=family_size, replace=False)
            family = Population.merge(pop[rnd], off)
            pop[rnd] = FitnessSurvival().do(self.problem, family, n_survive=family_size)

            for i in rnd:
                if is_better(pop[i], pop[0]):
                    tmp = pop[0]
                    pop[0] = pop[i]
                    pop[i] = tmp

    def _set_optimum(self, **kwargs):
        self.opt = self.pop[[0]]


parse_doc_string(G3PCX.__init__)



import numpy as np

from pymoo.core.mutation import Mutation


class BitflipMutation(Mutation):

    def _do(self, problem, X, **kwargs):
        prob_var = self.get_prob_var(problem, size=(len(X), 1))
        Xp = np.copy(X)
        flip = np.random.random(X.shape) < prob_var
        Xp[flip] = ~X[flip]
        return Xp


class BFM(BitflipMutation):
    pass



import math

import numpy as np

from pymoo.core.crossover import Crossover
from pymoo.util.misc import crossover_mask


class HalfUniformCrossover(Crossover):

    def __init__(self, prob_hux=0.5, **kwargs):
        super().__init__(2, 2, **kwargs)
        self.prob_hux = prob_hux

    def _do(self, _, X, **kwargs):
        _, n_matings, n_var = X.shape

        # the mask do to the crossover
        M = np.full((n_matings, n_var), False)

        not_equal = X[0] != X[1]

        # create for each individual the crossover range
        for i in range(n_matings):
            I = np.where(not_equal[i])[0]

            n = math.ceil(len(I) / 2)
            if n > 0:
                _I = I[np.random.permutation(len(I))[:n]]
                M[i, _I] = True

        _X = crossover_mask(X, M)
        return _X


class HUX(HalfUniformCrossover):
    pass



import numpy as np


def get_relation(ind_a, ind_b):
    return Dominator.get_relation(ind_a.F, ind_b.F, ind_a.CV[0], ind_b.CV[0])


class Dominator:

    @staticmethod
    def get_relation(a, b, cva=None, cvb=None):

        if cva is not None and cvb is not None:
            if cva < cvb:
                return 1
            elif cvb < cva:
                return -1

        val = 0
        for i in range(len(a)):
            if a[i] < b[i]:
                # indifferent because once better and once worse
                if val == -1:
                    return 0
                val = 1
            elif b[i] < a[i]:
                # indifferent because once better and once worse
                if val == 1:
                    return 0
                val = -1
        return val

    @staticmethod
    def calc_domination_matrix_loop(F, G):
        n = F.shape[0]
        CV = np.sum(G * (G > 0).astype(float), axis=1)
        M = np.zeros((n, n))
        for i in range(n):
            for j in range(i + 1, n):
                M[i, j] = Dominator.get_relation(F[i, :], F[j, :], CV[i], CV[j])
                M[j, i] = -M[i, j]

        return M

    @staticmethod
    def calc_domination_matrix(F, _F=None, epsilon=0.0):

        if _F is None:
            _F = F

        # look at the obj for dom
        n = F.shape[0]
        m = _F.shape[0]

        L = np.repeat(F, m, axis=0)
        R = np.tile(_F, (n, 1))

        smaller = np.reshape(np.any(L + epsilon < R, axis=1), (n, m))
        larger = np.reshape(np.any(L > R + epsilon, axis=1), (n, m))

        M = np.logical_and(smaller, np.logical_not(larger)) * 1 \
            + np.logical_and(larger, np.logical_not(smaller)) * -1

        # if cv equal then look at dom
        # M = constr + (constr == 0) * dom

        return M



import math

import numpy as np
from scipy.spatial.distance import cdist, pdist, squareform

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.core.population import Population
from pymoo.decomposition.asf import ASF
from pymoo.docs import parse_doc_string
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.selection.tournament import TournamentSelection
from pymoo.util.display.multi import MultiObjectiveOutput
from pymoo.util.dominator import Dominator
from pymoo.util.function_loader import load_function
from pymoo.util.misc import has_feasible, random_permuations
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting


# =========================================================================================================
# Implementation
# Following original code by K. Li https://cola-laboratory.github.io/codes/CTAEA.zip
# =========================================================================================================


def comp_by_cv_dom_then_random(pop, P, **kwargs):
    S = np.full(P.shape[0], np.nan)

    for i in range(P.shape[0]):
        a, b = P[i, 0], P[i, 1]

        if pop[a].CV <= 0.0 and pop[b].CV <= 0.0:
            rel = Dominator.get_relation(pop[a].F, pop[b].F)
            if rel == 1:
                S[i] = a
            elif rel == -1:
                S[i] = b
            else:
                S[i] = np.random.choice([a, b])
        elif pop[a].CV <= 0.0:
            S[i] = a
        elif pop[b].CV <= 0.0:
            S[i] = b
        else:
            S[i] = np.random.choice([a, b])

    return S[:, None].astype(int)


class RestrictedMating(TournamentSelection):
    """Restricted mating approach to balance convergence and diversity archives"""

    def _do(self, problem, Hm, n_select, n_parents, **kwargs):
        n_pop = len(Hm) // 2

        _, rank = NonDominatedSorting().do(Hm.get('F'), return_rank=True)

        Pc = (rank[:n_pop] == 0).sum() / len(Hm)
        Pd = (rank[n_pop:] == 0).sum() / len(Hm)

        # number of random individuals needed
        n_random = n_select * n_parents * self.pressure
        n_perms = math.ceil(n_random / n_pop)
        # get random permutations and reshape them
        P = random_permuations(n_perms, n_pop)[:n_random]
        P = np.reshape(P, (n_select * n_parents, self.pressure))
        if Pc <= Pd:
            # Choose from DA
            P[::n_parents, :] += n_pop
        pf = np.random.random(n_select)
        P[1::n_parents, :][pf >= Pc] += n_pop

        # compare using tournament function
        S = self.func_comp(Hm, P, **kwargs)

        return np.reshape(S, (n_select, n_parents))


class CADASurvival:

    def __init__(self, ref_dirs):
        self.ref_dirs = ref_dirs
        self.opt = None
        self.ideal_point = np.full(ref_dirs.shape[1], np.inf)
        self._decomposition = ASF()
        self._calc_perpendicular_distance = load_function("calc_perpendicular_distance")

    def do(self, _, pop, da, n_survive=None, **kwargs):
        # Offspring are last of merged population
        off = pop[-n_survive:]
        # Update ideal point
        self.ideal_point = np.min(np.vstack((self.ideal_point, off.get("F"))), axis=0)
        # Update CA
        pop = self._updateCA(pop, n_survive)
        # Update DA
        Hd = Population.merge(da, off)
        da = self._updateDA(pop, Hd, n_survive)
        return pop, da

    def _associate(self, pop):
        """Associate each individual with a F vector and calculate decomposed fitness"""
        F = pop.get("F")
        dist_matrix = self._calc_perpendicular_distance(F - self.ideal_point, self.ref_dirs)
        niche_of_individuals = np.argmin(dist_matrix, axis=1)
        FV = self._decomposition.do(F, weights=self.ref_dirs[niche_of_individuals, :],
                                    ideal_point=self.ideal_point, weight_0=1e-4)
        pop.set("niche", niche_of_individuals)
        pop.set("FV", FV)
        return niche_of_individuals, FV

    def _updateCA(self, pop, n_survive):
        """Update the Convergence archive (CA)"""
        CV = pop.get("CV").flatten()

        Sc = pop[CV == 0]  # ConstraintsAsObjective population
        if len(Sc) == n_survive:  # Exactly n_survive feasible individuals
            F = Sc.get("F")
            fronts, rank = NonDominatedSorting().do(F, return_rank=True)
            Sc.set('rank', rank)
            self.opt = Sc[fronts[0]]
            return Sc
        elif len(Sc) < n_survive:  # Not enough feasible individuals
            remainder = n_survive - len(Sc)
            # Solve sub-problem CV, tche
            SI = pop[CV > 0]
            f1 = SI.get("CV")
            _, f2 = self._associate(SI)
            sub_F = np.column_stack([f1, f2])
            fronts = NonDominatedSorting().do(sub_F, n_stop_if_ranked=remainder)
            I = []
            for front in fronts:
                if len(I) + len(front) <= remainder:
                    I.extend(front)
                else:
                    n_missing = remainder - len(I)
                    last_front_CV = np.argsort(f1.flatten()[front])
                    I.extend(front[last_front_CV[:n_missing]])
            SI = SI[I]
            S = Population.merge(Sc, SI)
            F = S.get("F")
            fronts, rank = NonDominatedSorting().do(F, return_rank=True)
            S.set('rank', rank)
            self.opt = S[fronts[0]]
            return S
        else:  # Too many feasible individuals
            F = Sc.get("F")
            # Filter by non-dominated sorting
            fronts, rank = NonDominatedSorting().do(F, return_rank=True, n_stop_if_ranked=n_survive)
            I = np.concatenate(fronts)
            S, rank, F = Sc[I], rank[I], F[I]
            if len(S) > n_survive:
                # Remove individual in most crowded niche and with worst fitness
                niche_of_individuals, FV = self._associate(S)
                index, count = np.unique(niche_of_individuals, return_counts=True)
                survivors = np.full(S.shape[0], True)
                while survivors.sum() > n_survive:
                    crowdest_niches, = np.where(count == count.max())
                    worst_idx = None
                    worst_niche = None
                    worst_fit = -1
                    for crowdest_niche in crowdest_niches:
                        crowdest, = np.where((niche_of_individuals == index[crowdest_niche]) & survivors)
                        niche_worst = crowdest[FV[crowdest].argmax()]
                        dist_to_max_fit = cdist(F[[niche_worst], :], F).flatten()
                        dist_to_max_fit[niche_worst] = np.inf
                        dist_to_max_fit[~survivors] = np.inf
                        min_d_to_max_fit = dist_to_max_fit.min()

                        dist_in_niche = squareform(pdist(F[crowdest]))
                        np.fill_diagonal(dist_in_niche, np.inf)

                        delta_d = dist_in_niche - min_d_to_max_fit
                        min_d_i = np.unravel_index(np.argmin(delta_d, axis=None), dist_in_niche.shape)
                        if (delta_d[min_d_i] < 0) or (
                                delta_d[min_d_i] == 0 and (FV[crowdest[list(min_d_i)]] > niche_worst).any()):
                            min_d_i = list(min_d_i)
                            np.random.shuffle(min_d_i)
                            closest = crowdest[min_d_i]
                            niche_worst = closest[np.argmax(FV[closest])]
                        if FV[niche_worst] > worst_fit:
                            worst_fit = FV[niche_worst]
                            worst_idx = niche_worst
                            worst_niche = crowdest_niche
                    survivors[worst_idx] = False
                    count[worst_niche] -= 1
                S, rank = S[survivors], rank[survivors]
            S.set('rank', rank)
            self.opt = S[rank == 0]
            return S

    def _updateDA(self, pop, Hd, n_survive):
        """Update the Diversity archive (DA)"""
        niche_Hd, FV = self._associate(Hd)
        niche_CA, _ = self._associate(pop)

        itr = 1
        S = []
        while len(S) < n_survive:
            for i in range(n_survive):
                current_ca, = np.where(niche_CA == i)
                if len(current_ca) < itr:
                    for _ in range(itr - len(current_ca)):
                        current_da = np.where(niche_Hd == i)[0]
                        if current_da.size > 0:
                            F = Hd[current_da].get('F')
                            nd = NonDominatedSorting().do(F, only_non_dominated_front=True, n_stop_if_ranked=0)
                            i_best = current_da[nd[np.argmin(FV[current_da[nd]])]]
                            niche_Hd[i_best] = -1
                            if len(S) < n_survive:
                                S.append(i_best)
                        else:
                            break
                if len(S) == n_survive:
                    break
            itr += 1
        return Hd[S]


class CTAEA(GeneticAlgorithm):

    def __init__(self,
                 ref_dirs,
                 sampling=FloatRandomSampling(),
                 selection=RestrictedMating(func_comp=comp_by_cv_dom_then_random),
                 crossover=SBX(n_offsprings=1, eta=30, prob=1.0),
                 mutation=PM(prob_var=None, eta=20),
                 eliminate_duplicates=True,
                 output=MultiObjectiveOutput(),
                 **kwargs):
        """
        CTAEA

        Parameters
        ----------
        ref_dirs : {ref_dirs}
        sampling : {sampling}
        selection : {selection}
        crossover : {crossover}
        mutation : {mutation}
        eliminate_duplicates : {eliminate_duplicates}
        """

        self.ref_dirs = ref_dirs
        pop_size = len(ref_dirs)

        if 'survival' in kwargs:
            survival = kwargs['survival']
            del kwargs['survival']
        else:
            survival = CADASurvival(ref_dirs)

        # Initialize diversity archives
        self.da = None

        super().__init__(pop_size=pop_size,
                         sampling=sampling,
                         selection=selection,
                         crossover=crossover,
                         mutation=mutation,
                         survival=survival,
                         eliminate_duplicates=eliminate_duplicates,
                         n_offsprings=pop_size,
                         output=output,
                         **kwargs)

    def _setup(self, problem, **kwargs):

        if self.ref_dirs is not None and self.ref_dirs.shape[1] != problem.n_obj:
            raise Exception(
                "Dimensionality of reference points must be equal to the number of objectives: %s != %s" %
                (self.ref_dirs.shape[1], problem.n_obj))

    def _initialize_infill(self):
        return self.initialization.do(self.problem, self.pop_size, algorithm=self)

    def _initialize_advance(self, infills=None, **kwargs):
        super()._initialize_advance(infills, **kwargs)
        self.pop, self.da = self.survival.do(self.problem, self.pop, Population(), n_survive=len(self.pop),
                                             algorithm=self)

    def _infill(self):
        Hm = Population.merge(self.pop, self.da)
        return self.mating.do(self.problem, Hm, n_offsprings=self.n_offsprings, algorithm=self)

    def _advance(self, infills=None, **kwargs):
        assert infills is not None, "This algorithms uses the AskAndTell interface thus infills must to be provided."
        pop = Population.merge(self.pop, infills)
        self.pop, self.da = self.survival.do(self.problem, pop, self.da, self.pop_size, algorithm=self)

    def _set_optimum(self, **kwargs):
        if not has_feasible(self.pop):
            self.opt = self.pop[[np.argmin(self.pop.get("CV"))]]
        else:
            self.opt = self.survival.opt


parse_doc_string(CTAEA.__init__)



from pymoo.indicators.distance_indicator import DistanceIndicator, modified_distance


class GDPlus(DistanceIndicator):

    def __init__(self, pf, **kwargs):
        super().__init__(pf, modified_distance, 0, **kwargs)



class Result:
    """
    The resulting object of an optimization run.
    """

    def __init__(self) -> None:
        super().__init__()

        self.opt = None
        self.success = None
        self.message = None

        # ! other attributes to be set as well

        # the problem that was solved
        self.problem = None

        # the archive stored during the run
        self.archive = None

        # the optimal solution for that problem
        self.pf = None

        # the algorithm that was used for optimization
        self.algorithm = None

        # the final population if it applies
        self.pop = None

        # directly the values of opt
        self.X, self.F, self.CV, self.G, self.H = None, None, None, None, None

        # all the timings that are stored of the run
        self.start_time, self.end_time, self.exec_time = None, None, None

        # the history of the optimization run is they were saved
        self.history = []

    @property
    def cv(self):
        return self.CV[0]

    @property
    def f(self):
        return self.F[0]

    @property
    def feas(self):
        return self.cv <= 0



import pymoo.gradient.toolbox as anp
import numpy as np

from pymoo.core.problem import Problem
from pymoo.util.reference_direction import UniformReferenceDirectionFactory
from pymoo.util.remote import Remote


class DTLZ(Problem):
    def __init__(self, n_var, n_obj, k=None, **kwargs):

        if n_var:
            self.k = n_var - n_obj + 1
        elif k:
            self.k = k
            n_var = k + n_obj - 1
        else:
            raise Exception("Either provide number of variables or k!")

        super().__init__(n_var=n_var, n_obj=n_obj, xl=0, xu=1, vtype=float, **kwargs)

    def g1(self, X_M):
        return 100 * (self.k + anp.sum(anp.square(X_M - 0.5) - anp.cos(20 * anp.pi * (X_M - 0.5)), axis=1))

    def g2(self, X_M):
        return anp.sum(anp.square(X_M - 0.5), axis=1)

    def obj_func(self, X_, g, alpha=1):
        f = []

        for i in range(0, self.n_obj):
            _f = (1 + g)
            _f *= anp.prod(anp.cos(anp.power(X_[:, :X_.shape[1] - i], alpha) * anp.pi / 2.0), axis=1)
            if i > 0:
                _f *= anp.sin(anp.power(X_[:, X_.shape[1] - i], alpha) * anp.pi / 2.0)

            f.append(_f)

        f = anp.column_stack(f)
        return f


class DTLZ1(DTLZ):
    def __init__(self, n_var=7, n_obj=3, **kwargs):
        super().__init__(n_var=n_var, n_obj=n_obj, **kwargs)

    def _calc_pareto_front(self, ref_dirs=None):
        if ref_dirs is None:
            ref_dirs = get_ref_dirs(self.n_obj)
        return 0.5 * ref_dirs

    def obj_func(self, X_, g):
        f = []

        for i in range(0, self.n_obj):
            _f = 0.5 * (1 + g)
            _f *= anp.prod(X_[:, :X_.shape[1] - i], axis=1)
            if i > 0:
                _f *= 1 - X_[:, X_.shape[1] - i]
            f.append(_f)

        return anp.column_stack(f)

    def _evaluate(self, x, out, *args, **kwargs):
        X_, X_M = x[:, :self.n_obj - 1], x[:, self.n_obj - 1:]
        g = self.g1(X_M)
        out["F"] = self.obj_func(X_, g)


class DTLZ2(DTLZ):
    def __init__(self, n_var=10, n_obj=3, **kwargs):
        super().__init__(n_var=n_var, n_obj=n_obj, **kwargs)

    def _calc_pareto_front(self, ref_dirs=None):
        if ref_dirs is None:
            ref_dirs = get_ref_dirs(self.n_obj)
        return generic_sphere(ref_dirs)

    def _evaluate(self, x, out, *args, **kwargs):
        X_, X_M = x[:, :self.n_obj - 1], x[:, self.n_obj - 1:]
        g = self.g2(X_M)
        out["F"] = self.obj_func(X_, g, alpha=1)


class DTLZ3(DTLZ):
    def __init__(self, n_var=10, n_obj=3, **kwargs):
        super().__init__(n_var=n_var, n_obj=n_obj, **kwargs)

    def _calc_pareto_front(self, ref_dirs=None):
        if ref_dirs is None:
            ref_dirs = get_ref_dirs(self.n_obj)
        return generic_sphere(ref_dirs)

    def _evaluate(self, x, out, *args, **kwargs):
        X_, X_M = x[:, :self.n_obj - 1], x[:, self.n_obj - 1:]
        g = self.g1(X_M)
        out["F"] = self.obj_func(X_, g, alpha=1)


class DTLZ4(DTLZ):
    def __init__(self, n_var=10, n_obj=3, alpha=100, d=100, **kwargs):
        super().__init__(n_var=n_var, n_obj=n_obj, **kwargs)
        self.alpha = alpha
        self.d = d

    def _calc_pareto_front(self, ref_dirs=None):
        if ref_dirs is None:
            ref_dirs = get_ref_dirs(self.n_obj)
        return generic_sphere(ref_dirs)

    def _evaluate(self, x, out, *args, **kwargs):
        X_, X_M = x[:, :self.n_obj - 1], x[:, self.n_obj - 1:]
        g = self.g2(X_M)
        out["F"] = self.obj_func(X_, g, alpha=self.alpha)


class DTLZ5(DTLZ):
    def __init__(self, n_var=10, n_obj=3, **kwargs):
        super().__init__(n_var=n_var, n_obj=n_obj, **kwargs)

    def _calc_pareto_front(self):
        if self.n_obj == 3:
            return Remote.get_instance().load("pymoo", "pf", "dtlz5-3d.pf")
        else:
            raise Exception("Not implemented yet.")

    def _evaluate(self, x, out, *args, **kwargs):
        X_, X_M = x[:, :self.n_obj - 1], x[:, self.n_obj - 1:]
        g = self.g2(X_M)

        theta = 1 / (2 * (1 + g[:, None])) * (1 + 2 * g[:, None] * X_)
        theta = anp.column_stack([x[:, 0], theta[:, 1:]])

        out["F"] = self.obj_func(theta, g)


class DTLZ6(DTLZ):
    def __init__(self, n_var=10, n_obj=3, **kwargs):
        super().__init__(n_var=n_var, n_obj=n_obj, **kwargs)

    def _calc_pareto_front(self):
        if self.n_obj == 3:
            return Remote.get_instance().load("pymoo", "pf", "dtlz6-3d.pf")
        else:
            raise Exception("Not implemented yet.")

    def _evaluate(self, x, out, *args, **kwargs):
        X_, X_M = x[:, :self.n_obj - 1], x[:, self.n_obj - 1:]
        g = anp.sum(anp.power(X_M, 0.1), axis=1)

        theta = 1 / (2 * (1 + g[:, None])) * (1 + 2 * g[:, None] * X_)
        theta = anp.column_stack([x[:, 0], theta[:, 1:]])

        out["F"] = self.obj_func(theta, g)


class DTLZ7(DTLZ):
    def __init__(self, n_var=10, n_obj=3, **kwargs):
        super().__init__(n_var=n_var, n_obj=n_obj, **kwargs)

    def _calc_pareto_front(self):
        if self.n_obj == 3:
            return Remote.get_instance().load("pymoo", "pf", "dtlz7-3d.pf")
        else:
            raise Exception("Not implemented yet.")

    def _evaluate(self, x, out, *args, **kwargs):
        f = []
        for i in range(0, self.n_obj - 1):
            f.append(x[:, i])
        f = anp.column_stack(f)

        g = 1 + 9 / self.k * anp.sum(x[:, -self.k:], axis=1)
        h = self.n_obj - anp.sum(f / (1 + g[:, None]) * (1 + anp.sin(3 * anp.pi * f)), axis=1)

        out["F"] = anp.column_stack([f, (1 + g) * h])


class InvertedDTLZ1(DTLZ1):

    def _calc_pareto_front(self):
        raise Exception("Not implemented yet.")

    def _evaluate(self, x, out, *args, **kwargs):
        X_, X_M = x[:, :self.n_obj - 1], x[:, self.n_obj - 1:]
        g = self.g1(X_M)

        super()._evaluate(x, out, *args, **kwargs)
        out["F"] = 0.5 * (1 + g[:, None]) - out["F"]

    def _calc_pareto_front(self, *args, **kwargs):
        return self.problem.pareto_front(*args, **kwargs)


class ScaledProblem(Problem):

    def __init__(self, problem, scale_factor):
        super().__init__(n_var=problem.n_var, n_obj=problem.n_obj, n_ieq_constr=problem.n_ieq_constr,
                         n_eq_constr=problem.n_eq_constr, xl=problem.xl, xu=problem.xu, vtype=problem.vtype)
        self.problem = problem
        self.scale_factor = scale_factor

    @staticmethod
    def get_scale(n, scale_factor):
        return np.power(np.full(n, scale_factor), np.arange(n))

    def _evaluate(self, X, out, *args, **kwargs):
        self.problem._evaluate(X, out, *args, **kwargs)
        out["F"] = out["F"] * ScaledProblem.get_scale(self.n_obj, self.scale_factor)

    def _calc_pareto_front(self, *args, **kwargs):
        return self.problem.pareto_front(*args, **kwargs) * ScaledProblem.get_scale(self.n_obj, self.scale_factor)


class ConvexProblem(Problem):

    def __init__(self, problem):
        super().__init__(problem.n_var, problem.n_obj, problem.n_ieq_constr, problem.n_eq_constr, problem.xl, problem.xu)
        self.problem = problem

    def get_power(self, n):
        p = np.full(n, 4.0)
        p[-1] = 2.0
        return p

    def _evaluate(self, X, out, *args, **kwargs):
        self.problem._evaluate(X, out, **kwargs)
        out["F"] = anp.power(out["F"], self.get_power(self.n_obj))

    def _calc_pareto_front(self, ref_dirs, *args, **kwargs):
        F = self.problem.pareto_front(ref_dirs)
        return np.power(F, self.get_power(self.n_obj))


class ScaledDTLZ1(ScaledProblem):

    def __init__(self, n_var=7, n_obj=3, scale_factor=10, **kwargs):
        super().__init__(DTLZ1(n_var=n_var, n_obj=n_obj, **kwargs), scale_factor=scale_factor)


class ConvexDTLZ2(ConvexProblem):

    def __init__(self, n_var=10, n_obj=3, **kwargs):
        super().__init__(DTLZ2(n_var=n_var, n_obj=n_obj, **kwargs))


class ConvexDTLZ4(ConvexProblem):

    def __init__(self, n_var=10, n_obj=3, **kwargs):
        super().__init__(DTLZ4(n_var=n_var, n_obj=n_obj, **kwargs))


def generic_sphere(ref_dirs):
    return ref_dirs / np.tile(np.linalg.norm(ref_dirs, axis=1)[:, None], (1, ref_dirs.shape[1]))


def get_ref_dirs(n_obj):
    if n_obj == 2:
        ref_dirs = UniformReferenceDirectionFactory(2, n_points=100).do()
    elif n_obj == 3:
        ref_dirs = UniformReferenceDirectionFactory(3, n_partitions=15).do()
    else:
        raise Exception("Please provide reference directions for more than 3 objectives!")
    return ref_dirs



import sys
import importlib

GRADIENT_MODULE = "pymoo.gradient.toolbox"

TOOLBOX = "autograd.numpy"


def activate(name):
    global TOOLBOX
    TOOLBOX = name

    if GRADIENT_MODULE in sys.modules:
        del sys.modules[GRADIENT_MODULE]
    importlib.import_module(GRADIENT_MODULE)


def deactivate():
    activate("numpy")



import numpy as np

from pymoo.docs import parse_doc_string
from pymoo.core.plot import Plot
from pymoo.util.misc import all_combinations


class FitnessLandscape(Plot):

    def __init__(self,
                 problem,
                 _type="surface+contour",
                 n_samples=100,
                 colorbar=False,
                 contour_levels=30,
                 kwargs_surface=None,
                 kwargs_contour=None,
                 kwargs_contour_labels=None,
                 **kwargs):

        """

        Fitness Landscape

        Parameters
        ----------------

        problem : The problem to be plotted
        _type : str
            Either "contour", "surface" or "contour+surface"
        labels : {labels}

        Other Parameters
        ----------------

        figsize : {figsize}
        title : {title}
        legend : {legend}
        tight_layout : {tight_layout}
        cmap : {cmap}


        Returns
        -------
        FitnessLandscape : :class:`~pymoo.core.analytics.visualization.fitness_landscape.FitnessLandscape`

        """

        super().__init__(**kwargs)
        self.problem = problem
        self.n_samples = n_samples
        self._type = _type
        self.colorbar = colorbar

        self.contour_levels = contour_levels

        self.kwargs_surface = kwargs_surface
        if self.kwargs_surface is None:
            self.kwargs_surface = dict(cmap="summer", rstride=1, cstride=1)

        self.kwargs_contour = kwargs_contour
        if self.kwargs_contour is None:
            self.kwargs_contour = dict(linestyles="solid", offset=-1)

        self.kwargs_contour_labels = kwargs_contour_labels

    def _do(self):

        problem, n_samples, _type = self.problem, self.n_samples, self._type

        if problem.n_var == 1 and problem.n_obj == 1:

            self.init_figure()

            X = np.linspace(problem.xl[0], problem.xu[0], num=n_samples)[:, None]
            Z = problem.evaluate(X, return_values_of=["F"])
            self.ax.plot(X, Z)
            self.ax.set_xlabel("x")
            self.ax.set_ylabel("f(x)")

        elif problem.n_var == 2 and problem.n_obj == 1:

            A = np.linspace(problem.xl[0], problem.xu[0], n_samples)
            B = np.linspace(problem.xl[1], problem.xu[1], n_samples)
            X = all_combinations(A, B)

            F = np.reshape(problem.evaluate(X, return_values_of=["F"]), (n_samples, n_samples))

            _X = X[:, 0].reshape((n_samples, n_samples))
            _Y = X[:, 1].reshape((n_samples, n_samples))
            _Z = F.reshape((n_samples, n_samples))

            def plot_surface():
                surf = self.ax.plot_surface(_X, _Y, _Z, **self.kwargs_surface)

                if self.colorbar:
                    self.fig.colorbar(surf)

            def plot_contour():
                CS = self.ax.contour(_X, _Y, _Z, self.contour_levels, **self.kwargs_contour)
                if self.kwargs_contour_labels is not None:
                    self.ax.clabel(CS, **self.kwargs_contour_labels)

                if self.colorbar:
                    self.fig.colorbar(CS)

            if _type == "surface":
                self.init_figure(plot_3D=True)
                plot_surface()
            elif _type == "contour":
                self.init_figure(plot_3D=False)

                if "offset" in self.kwargs_contour:
                    del self.kwargs_contour["offset"]

                plot_contour()
            elif _type == "surface+contour":

                self.init_figure(plot_3D=True)
                plot_surface()
                plot_contour()

        else:
            raise Exception("Only landscapes of problems with one or two variables and one objective can be visualized.")


parse_doc_string(FitnessLandscape.__init__)



import numpy as np

from pymoo.docs import parse_doc_string
from pymoo.core.plot import Plot
from pymoo.util.misc import set_if_none, set_if_none_from_tuples
from pymoo.visualization.util import parse_bounds, normalize


class PCP(Plot):

    def __init__(self,
                 bounds=None,
                 show_bounds=True,
                 n_ticks=5,
                 normalize_each_axis=True,
                 bbox=False,
                 **kwargs):
        """

        Parallel Coordinate Plot


        Parameters
        ----------------

        bounds : {bounds}

        axis_style : {axis_style}

        labels : {labels}

        n_ticks : int
            Number of ticks to be shown on each parallel axis.

        show_bounds : bool
            Whether the value of the boundaries are shown in the plot or not.

        normalize_each_axis : bool
            Whether the values should be normalized either by bounds or implictly.

        Other Parameters
        ----------------

        figsize : {figsize}
        title : {title}
        legend : {legend}
        tight_layout : {tight_layout}
        cmap : {cmap}

        """

        super().__init__(bounds=bounds, **kwargs)
        self.show_bounds = show_bounds
        self.n_ticks = n_ticks
        self.bbox = bbox
        self.normalize_each_axis = normalize_each_axis

        set_if_none_from_tuples(self.axis_style, ("color", "red"), ("linewidth", 2), ("alpha", 0.75))

    def _do(self):

        # initial a figure with a single plot
        self.init_figure()

        # if no normalization of each axis the bounds are based on the overall min and max
        if not self.normalize_each_axis and self.bounds is None:
            _F = np.row_stack([e[0] for e in self.to_plot])
            self.bounds = [_F.min(), _F.max()]

        # normalize the input
        bounds = parse_bounds(self.bounds, self.n_dim)
        to_plot_norm, bounds = normalize(self.to_plot, bounds, return_bounds=True)

        # plot for each set the lines
        for k, (F, kwargs) in enumerate(to_plot_norm):

            _kwargs = kwargs.copy()
            set_if_none(_kwargs, "color", self.colors[k % len(self.colors)])

            for i in range(len(F)):
                self.ax.plot(np.arange(F.shape[1]), F[i, :], **_kwargs)

        # Plot the parallel coordinate axes
        for i in range(self.n_dim):
            self.ax.axvline(i, **self.axis_style)

            bottom, top = -0.1, 1.075
            margin_left = 0.08

            if self.show_bounds:
                lower = self.ax.text(i - margin_left, bottom, self.func_number_to_text(bounds[0][i]))
                upper = self.ax.text(i - margin_left, top, self.func_number_to_text(bounds[1][i]))

                if self.bbox:
                    lower.set_bbox(dict(facecolor='white', alpha=0.8))
                    upper.set_bbox(dict(facecolor='white', alpha=0.8))

            if self.n_ticks is not None:
                n_length = 0.03
                for y in np.linspace(0, 1, self.n_ticks):
                    self.ax.hlines(y, i - n_length, i + n_length, **self.axis_style)

        # if bounds are shown, then move them to the bottom
        if self.show_bounds:
            self.ax.tick_params(axis='x', which='major', pad=25)

        self.ax.spines['right'].set_visible(False)
        self.ax.spines['left'].set_visible(False)

        self.ax.set_yticklabels([])
        self.ax.set_yticks([])
        self.ax.set_ylim((-0.05, 1.05))

        self.ax.set_xticks(np.arange(self.n_dim))
        self.ax.set_xticklabels(self.get_labels())

        return self


parse_doc_string(PCP.__init__)



from abc import abstractmethod

import numpy as np

import pymoo.gradient.toolbox as anp
from pymoo.util.cache import Cache
from pymoo.util.misc import at_least_2d_array


class ElementwiseEvaluationFunction:

    def __init__(self, problem, args, kwargs) -> None:
        super().__init__()
        self.problem = problem
        self.args = args
        self.kwargs = kwargs

    def __call__(self, x):
        out = dict()
        self.problem._evaluate(x, out, *self.args, **self.kwargs)
        return out


class LoopedElementwiseEvaluation:

    def __call__(self, f, X):
        return [f(x) for x in X]


class StarmapParallelization:

    def __init__(self, starmap) -> None:
        super().__init__()
        self.starmap = starmap

    def __call__(self, f, X):
        return list(self.starmap(f, [[x] for x in X]))

    def __getstate__(self):
        state = self.__dict__.copy()
        state.pop("starmap", None)
        return state


class DaskParallelization:

    def __init__(self, client) -> None:
        super().__init__()
        self.client = client

    def __call__(self, f, X):
        jobs = [self.client.submit(f, x) for x in X]
        return [job.result() for job in jobs]

    def __getstate__(self):
        state = self.__dict__.copy()
        state.pop("client")
        return state


class Problem:
    def __init__(self,
                 n_var=-1,
                 n_obj=1,
                 n_ieq_constr=0,
                 n_eq_constr=0,
                 xl=None,
                 xu=None,
                 vtype=None,
                 vars=None,
                 elementwise=False,
                 elementwise_func=ElementwiseEvaluationFunction,
                 elementwise_runner=LoopedElementwiseEvaluation(),
                 replace_nan_values_by=None,
                 exclude_from_serialization=None,
                 callback=None,
                 strict=True,
                 **kwargs):

        """

        Parameters
        ----------
        n_var : int
            Number of Variables

        n_obj : int
            Number of Objectives

        n_ieq_constr : int
            Number of Inequality Constraints

        n_eq_constr : int
            Number of Equality Constraints

        xl : np.array, float, int
            Lower bounds for the variables. if integer all lower bounds are equal.

        xu : np.array, float, int
            Upper bounds for the variable. if integer all upper bounds are equal.

        vtype : type
            The variable type. So far, just used as a type hint.

        """

        # if variables are provided directly
        if vars is not None:
            n_var = len(vars)

        # number of variable
        self.n_var = n_var

        # number of objectives
        self.n_obj = n_obj

        # number of inequality constraints
        self.n_ieq_constr = n_ieq_constr if "n_constr" not in kwargs else max(n_ieq_constr, kwargs["n_constr"])

        # number of equality constraints
        self.n_eq_constr = n_eq_constr

        # type of the variable to be evaluated
        self.data = dict(**kwargs)

        # the lower bounds, make sure it is a numpy array with the length of n_var
        self.xl, self.xu = xl, xu

        # a callback function to be called after every evaluation
        self.callback = callback

        # if the variables are provided in their explicit form
        self.vars = vars

        # the variable type (only as a type hint at this point)
        self.vtype = vtype

        # the functions used if elementwise is enabled
        self.elementwise = elementwise
        self.elementwise_func = elementwise_func
        self.elementwise_runner = elementwise_runner

        # whether the shapes are checked strictly
        self.strict = strict

        # if it is a problem with an actual number of variables - make sure xl and xu are numpy arrays
        if n_var > 0:

            if self.xl is not None:
                if not isinstance(self.xl, np.ndarray):
                    self.xl = np.ones(n_var) * xl
                self.xl = self.xl.astype(float)

            if self.xu is not None:
                if not isinstance(self.xu, np.ndarray):
                    self.xu = np.ones(n_var) * xu
                self.xu = self.xu.astype(float)

        # this defines if NaN values should be replaced or not
        self.replace_nan_values_by = replace_nan_values_by

        # attribute which are excluded from being serialized
        self.exclude_from_serialization = exclude_from_serialization

    def evaluate(self,
                 X,
                 *args,
                 return_values_of=None,
                 return_as_dictionary=False,
                 **kwargs):

        if return_values_of is None:
            return_values_of = ["F"]
            if self.n_ieq_constr > 0:
                return_values_of.append("G")
            if self.n_eq_constr > 0:
                return_values_of.append("H")

        # make sure the array is at least 2d. store if reshaping was necessary
        if isinstance(X, np.ndarray) and X.dtype != object:
            X, only_single_value = at_least_2d_array(X, extend_as="row", return_if_reshaped=True)
            assert X.shape[1] == self.n_var, f'Input dimension {X.shape[1]} are not equal to n_var {self.n_var}!'
        else:
            only_single_value = not (isinstance(X, list) or isinstance(X, np.ndarray))

        # this is where the actual evaluation takes place
        _out = self.do(X, return_values_of, *args, **kwargs)

        out = {}
        for k, v in _out.items():

            # copy it to a numpy array (it might be one of jax at this point)
            v = np.array(v)

            # in case the input had only one dimension, then remove always the first dimension from each output
            if only_single_value:
                v = v[0]

            # if the NaN values should be replaced
            if self.replace_nan_values_by is not None:
                v[np.isnan(v)] = self.replace_nan_values_by

            try:
                out[k] = v.astype(np.float64)
            except:
                pass

        if self.callback is not None:
            self.callback(X, out)

        # now depending on what should be returned prepare the output
        if return_as_dictionary:
            return out

        if len(return_values_of) == 1:
            return out[return_values_of[0]]
        else:
            return tuple([out[e] for e in return_values_of])

    def do(self, X, return_values_of, *args, **kwargs):

        # create an empty dictionary
        out = {name: None for name in return_values_of}

        # do the function evaluation
        if self.elementwise:
            self._evaluate_elementwise(X, out, *args, **kwargs)
        else:
            self._evaluate_vectorized(X, out, *args, **kwargs)

        # finally format the output dictionary
        out = self._format_dict(out, len(X), return_values_of)

        return out

    def _evaluate_vectorized(self, X, out, *args, **kwargs):
        self._evaluate(X, out, *args, **kwargs)

    def _evaluate_elementwise(self, X, out, *args, **kwargs):

        # create the function that evaluates a single individual
        f = self.elementwise_func(self, args, kwargs)

        # execute the runner
        elems = self.elementwise_runner(f, X)

        # for each evaluation call
        for elem in elems:

            # for each key stored for this evaluation
            for k, v in elem.items():

                # if the element does not exist in out yet -> create it
                if out.get(k, None) is None:
                    out[k] = []

                out[k].append(v)

        # convert to arrays (the none check is important because otherwise an empty array is initialized)
        for k in out:
            if out[k] is not None:
                out[k] = anp.array(out[k])

    def _format_dict(self, out, N, return_values_of):

        # get the default output shape for the default values
        shape = default_shape(self, N)

        # finally the array to be returned
        ret = {}

        # for all values that have been set in the user implemented function
        for name, v in out.items():

            # only if they have truly been set
            if v is not None:

                # if there is a shape to be expected
                if name in shape:

                    if isinstance(v, list):
                        v = anp.column_stack(v)

                    try:
                        v = v.reshape(shape[name])
                    except Exception as e:
                        raise Exception(
                            f"Problem Error: {name} can not be set, expected shape {shape[name]} but provided {v.shape}", e)

                ret[name] = v

        # if some values that are necessary have not been set
        for name in return_values_of:
            if name not in ret:
                s = shape.get(name, N)
                ret[name] = np.full(s, np.inf)

        return ret

    @Cache
    def nadir_point(self, *args, **kwargs):
        pf = self.pareto_front(*args, **kwargs)
        if pf is not None:
            return np.max(pf, axis=0)

    @Cache
    def ideal_point(self, *args, **kwargs):
        pf = self.pareto_front(*args, **kwargs)
        if pf is not None:
            return np.min(pf, axis=0)

    @Cache
    def pareto_front(self, *args, **kwargs):
        pf = self._calc_pareto_front(*args, **kwargs)
        pf = at_least_2d_array(pf, extend_as='r')
        if pf is not None and pf.shape[1] == 2:
            pf = pf[np.argsort(pf[:, 0])]
        return pf

    @Cache
    def pareto_set(self, *args, **kwargs):
        ps = self._calc_pareto_set(*args, **kwargs)
        ps = at_least_2d_array(ps, extend_as='r')
        return ps

    @property
    def n_constr(self):
        return self.n_ieq_constr + self.n_eq_constr

    @abstractmethod
    def _evaluate(self, x, out, *args, **kwargs):
        pass

    def has_bounds(self):
        return self.xl is not None and self.xu is not None

    def has_constraints(self):
        return self.n_constr > 0

    def bounds(self):
        return self.xl, self.xu

    def name(self):
        return self.__class__.__name__

    def _calc_pareto_front(self, *args, **kwargs):
        pass

    def _calc_pareto_set(self, *args, **kwargs):
        pass

    def __str__(self):
        s = "# name: %s\n" % self.name()
        s += "# n_var: %s\n" % self.n_var
        s += "# n_obj: %s\n" % self.n_obj
        s += "# n_ieq_constr: %s\n" % self.n_ieq_constr
        s += "# n_eq_constr: %s\n" % self.n_eq_constr
        return s

    def __getstate__(self):
        if self.exclude_from_serialization is not None:
            state = self.__dict__.copy()

            # exclude objects which should not be stored
            for key in self.exclude_from_serialization:
                state[key] = None

            return state
        else:
            return self.__dict__


class ElementwiseProblem(Problem):

    def __init__(self, elementwise=True, **kwargs):
        super().__init__(elementwise=elementwise, **kwargs)


def default_shape(problem, n):
    n_var = problem.n_var
    DEFAULTS = dict(
        F=(n, problem.n_obj),
        G=(n, problem.n_ieq_constr),
        H=(n, problem.n_eq_constr),
        dF=(n, problem.n_obj, n_var),
        dG=(n, problem.n_ieq_constr, n_var),
        dH=(n, problem.n_eq_constr, n_var),
    )
    return DEFAULTS



import numpy as np

from pymoo.util.misc import vectorized_cdist
from pymoo.util.ref_dirs.misc import project_onto_sum_equals_zero_plane, project_onto_unit_simplex_recursive
from pymoo.util.ref_dirs.optimizer import Adam
from pymoo.util.reference_direction import ReferenceDirectionFactory, map_onto_unit_simplex


def calc_dist_to_others(x, X):
    return - np.sqrt(((x - X) ** 2).sum(axis=1)).min()


def calc_dist_to_others_with_gradient(x, X):
    diff = (x - X)
    D = np.sqrt((diff ** 2).sum(axis=1))

    k = D.argmin()

    obj = - D.min()
    grad = - diff[k] / D[k]

    return obj, grad


class ConstructionBasedReferenceDirectionFactory(ReferenceDirectionFactory):

    def __init__(self,
                 n_dim,
                 n_points,
                 n_samples=100,
                 gradient_descent=True,
                 verbose=False,
                 **kwargs):

        super().__init__(n_dim, **kwargs)
        self.n_points = n_points
        self.gradient_descent = gradient_descent
        self.n_samples = n_samples
        self.verbose = verbose
        self.X = None

    def _do(self):

        self.X = np.eye(self.n_dim)

        while len(self.X) < self.n_points:
            x = self.next()
            self.X = np.row_stack([self.X, x])

            if self.verbose:
                print(len(self.X), "x", x)

        return self.X

    def next(self):

        x = np.random.random((self.n_samples, self.n_dim))
        x = map_onto_unit_simplex(x, "kraemer")
        x = x[vectorized_cdist(x, self.X).min(axis=1).argmax()]

        if self.gradient_descent:

            optimizer = Adam(precision=1e-4)

            # for each iteration of gradient descent
            for i in range(1000):

                # calculate the function value and the gradient
                # auto_obj, auto_grad = value_and_grad(calc_dist_to_others)(x, self.X)
                _obj, _grad = calc_dist_to_others_with_gradient(x, self.X)

                # project the gradient to have a sum of zero - guarantees to stay on the simplex
                proj_grad = project_onto_sum_equals_zero_plane(_grad)

                # apply a step of gradient descent by subtracting the projected gradient with a learning rate
                x = optimizer.next(x, proj_grad)

                # project the out of bounds points back onto the unit simplex
                project_onto_unit_simplex_recursive(x[None, :])

                # because of floating point issues make sure it is on the unit simplex
                x /= x.sum()

                # if there was only a little movement during the last iteration -> terminate
                if optimizer.has_converged:
                    break

        return x



from pymoo.problems.single.ackley import *
from pymoo.problems.single.cantilevered_beam import *
from pymoo.problems.single.griewank import *
from pymoo.problems.single.knapsack import *
from pymoo.problems.single.pressure_vessel import *
from pymoo.problems.single.rastrigin import *
from pymoo.problems.single.rosenbrock import *
from pymoo.problems.single.schwefel import *
from pymoo.problems.single.sphere import *
from pymoo.problems.single.zakharov import *
from pymoo.problems.single.g import *
from pymoo.problems.single.himmelblau import *


import numpy as np

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.algorithms.moo.nsga3 import HyperplaneNormalization
from pymoo.core.survival import Survival
from pymoo.docs import parse_doc_string
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.selection.tournament import TournamentSelection, compare
from pymoo.termination.default import DefaultMultiObjectiveTermination
from pymoo.util.display.multi import MultiObjectiveOutput
from pymoo.util.dominator import Dominator
from pymoo.util.misc import vectorized_cdist


# ---------------------------------------------------------------------------------------------------------
# Environmental Survival (in the original paper it is referred to as archiving)
# ---------------------------------------------------------------------------------------------------------


class SPEA2Survival(Survival):

    def __init__(self, normalize=False, filter_infeasible=True):
        super().__init__(filter_infeasible)

        # whether the survival should considered normalized distance or just raw
        self.normalize = normalize

        # an object keeping track of normalization points
        self.norm = None

    def _do(self, problem, pop, *args, n_survive=None, **kwargs):

        # get the objective space values and objects
        F = pop.get("F").astype(float, copy=False)

        # the domination matrix to see the relation for each solution to another
        M = Dominator().calc_domination_matrix(F)

        # the number of solutions each individual dominates
        S = (M == 1).sum(axis=0)

        # the raw fitness of each solution - strength of its dominators
        R = ((M == -1) * S).sum(axis=1)

        # determine what k-th nearest neighbor to consider
        k = int(np.sqrt(len(pop)))
        if k >= len(pop):
            k = len(pop) - 1

        # if normalization is enabled keep track of ideal and nadir
        if self.normalize:

            # initialize the first time and then always update the boundary points
            if self.norm is None:
                self.norm = HyperplaneNormalization(F.shape[1])
            self.norm.update(F)

            ideal, nadir = self.norm.ideal_point, self.norm.nadir_point

            _F = (F - ideal) / (nadir - ideal)
            dists = vectorized_cdist(_F, _F, fill_diag_with_inf=True)

        # if no normalize is required simply use the F values from the population
        else:
            dists = vectorized_cdist(F, F, fill_diag_with_inf=True)

        # the distances sorted for each individual
        sdists = np.sort(dists, axis=1)

        # inverse distance as part of the fitness
        D = 1 / (sdists[:, k] + 2)

        # the actual fitness value used to determine who survives
        SPEA_F = R + D

        # set all the attributes to the population
        pop.set(SPEA_F=SPEA_F, SPEA_R=R, SPEA_D=D)

        # get all the non-dominated solutions
        survivors = list(np.where(np.all(M >= 0, axis=1))[0])

        # if we normalize give boundary points most importance - give the boundary points in the nds set the lowest fit.
        if self.normalize:
            I = vectorized_cdist(self.norm.extreme_points, F[survivors]).argmin(axis=1)
            pop[survivors][I].set("SPEA_F", -1.0)

        # identify the remaining individuals to choose from
        H = set(survivors)
        rem = np.array([k for k in range(len(pop)) if k not in H])

        # if not enough solutions, will up by F
        if len(survivors) < n_survive:

            # sort them by the fitness values (lower is better) and append them
            rem_by_F = rem[SPEA_F[rem].argsort()]
            survivors.extend(rem_by_F[:n_survive - len(survivors)])

        # if too many, delete based on distances
        elif len(survivors) > n_survive:

            # remove one individual per loop, until we hit n_survive
            while len(survivors) > n_survive:
                i = dists[survivors][:, survivors].min(axis=1).argmin()
                survivors = [survivors[j] for j in range(len(survivors)) if j != i]

        return pop[survivors]


# ---------------------------------------------------------------------------------------------------------
# Binary Tournament Selection
# ---------------------------------------------------------------------------------------------------------


def spea_binary_tournament(pop, P, algorithm, **kwargs):
    n_tournaments, n_parents = P.shape

    if n_parents != 2:
        raise ValueError("Only implemented for binary tournament!")

    S = np.full(n_tournaments, np.nan)

    for i in range(n_tournaments):

        a, b = P[i, 0], P[i, 1]
        a_cv, a_f, b_cv, b_f, = pop[a].CV[0], pop[a].F, pop[b].CV[0], pop[b].F

        # if at least one solution is infeasible
        if a_cv > 0.0 or b_cv > 0.0:
            S[i] = compare(a, a_cv, b, b_cv, method='smaller_is_better', return_random_if_equal=True)

        # both solutions are feasible
        else:
            S[i] = compare(a, pop[a].get("SPEA_F"), b, pop[b].get("SPEA_F"), method='smaller_is_better',
                           return_random_if_equal=True)

    return S[:, None].astype(int, copy=False)


# ---------------------------------------------------------------------------------------------------------
# Algorithm
# ---------------------------------------------------------------------------------------------------------


class SPEA2(GeneticAlgorithm):

    def __init__(self,
                 pop_size=100,
                 sampling=FloatRandomSampling(),
                 selection=TournamentSelection(spea_binary_tournament),
                 crossover=SBX(),
                 mutation=PM(),
                 survival=SPEA2Survival(normalize=True),
                 eliminate_duplicates=True,
                 n_offsprings=None,
                 output=MultiObjectiveOutput(),
                 **kwargs):
        """

        SPEA2 - Strength Pareto EA 2

        Parameters
        ----------
        pop_size : {pop_size}
        sampling : {sampling}
        selection : {selection}
        crossover : {crossover}
        mutation : {mutation}
        eliminate_duplicates : {eliminate_duplicates}
        n_offsprings : {n_offsprings}

        """

        super().__init__(pop_size=pop_size,
                         sampling=sampling,
                         selection=selection,
                         crossover=crossover,
                         mutation=mutation,
                         survival=survival,
                         eliminate_duplicates=eliminate_duplicates,
                         n_offsprings=n_offsprings,
                         output=output,
                         advance_after_initial_infill=True,
                         **kwargs)
        self.termination = DefaultMultiObjectiveTermination()


parse_doc_string(SPEA2.__init__)




import numpy as np


class Optimizer:

    def __init__(self, precision=1e-6) -> None:
        super().__init__()
        self.has_converged = False
        self.precision = precision

    def next(self, X, dX):
        _X = self._next(X, dX)

        if np.abs(_X - X).mean() < self.precision:
            self.has_converged = True

        return _X


class GradientDescent(Optimizer):

    def __init__(self, learning_rate=0.01, **kwargs) -> None:
        super().__init__(**kwargs)
        self.learning_rate = learning_rate

    def _next(self, X, dX):
        return X - self.learning_rate * dX


class Adam(Optimizer):

    def __init__(self, alpha=0.01, beta_1=0.9, beta_2=0.999, epsilon=1e-8, **kwargs) -> None:
        super().__init__(**kwargs)

        self.alpha = alpha
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon

        self.m_t = 0
        self.v_t = 0
        self.t = 0

    def _next(self, X, dX):
        self.t += 1
        beta_1, beta_2 = self.beta_1, self.beta_2

        # update moving average of gradient and squared gradient
        self.m_t = beta_1 * self.m_t + (1 - beta_1) * dX
        self.v_t = beta_2 * self.v_t + (1 - beta_2) * (dX * dX)

        # calculates the bias-corrected estimates
        m_cap = self.m_t / (1 - (beta_1 ** self.t))
        v_cap = self.v_t / (1 - (beta_2 ** self.t))

        # do the gradient update
        _X = X - (self.alpha * m_cap) / (np.sqrt(v_cap) + self.epsilon)

        return _X



import numpy as np


class Column:

    def __init__(self, name, width=13, func=None, truncate=True) -> None:
        super().__init__()
        self.name = name
        self.func = func
        self.width = width
        self.truncate = truncate
        self.value = None

    def update(self, algorithm):
        if self.func:
            self.value = self.func(algorithm)

    def header(self):
        return str(self.name).center(self.width)

    def text(self):
        value = self.value
        if value is None:
            value = "-"

        return format_text(value, self.width, self.truncate)

    def set(self, value):
        self.value = value


def number_to_text(number, width):
    if number >= 10 or number * 1e5 < 1:
        return f"%.{width - 7}E" % number
    else:
        return f"%.{width - 3}f" % number


def format_text(value, width, truncate):
    if value is not None:

        if isinstance(value, (np.floating, float)):
            text = number_to_text(value, width)
        else:
            text = str(value)

        if truncate and len(text) > width:
            text = text[:width]
    else:
        text = "-"
    text = text.rjust(width)
    return text



import numpy as np
import pymoo.gradient.toolbox as anp
from pymoo.core.problem import Problem


class Griewank(Problem):
    def __init__(self, n_var=2):
        super().__init__(n_var=n_var, n_obj=1, xl=-600, xu=600, vtype=float)

    def _evaluate(self, x, out, *args, **kwargs):
        out["F"] = 1 + 1 / 4000 * anp.sum(anp.power(x, 2), axis=1) \
                  - anp.prod(anp.cos(x / anp.sqrt(anp.arange(1, x.shape[1] + 1))), axis=1)

    def _calc_pareto_front(self):
        return 0

    def _calc_pareto_set(self):
        return np.full(self.n_var, 0)



import numpy as np

from pymoo.core.indicator import Indicator
from pymoo.util.misc import vectorized_cdist, at_least_2d_array


def euclidean_distance(a, b, norm=None):
    return np.sqrt((((a - b) / norm) ** 2).sum(axis=1))


def modified_distance(z, a, norm=None):
    d = a - z
    d[d < 0] = 0
    d = d / norm
    return np.sqrt((d ** 2).sum(axis=1))


def derive_ideal_and_nadir_from_pf(pf, ideal=None, nadir=None):

    # try to derive ideal and nadir if not already set and pf provided
    if pf is not None:
        if ideal is None:
            ideal = np.min(pf, axis=0)
        if nadir is None:
            nadir = np.max(pf, axis=0)

    return ideal, nadir


class DistanceIndicator(Indicator):

    def __init__(self, pf, dist_func, axis, zero_to_one=False, ideal=None, nadir=None, norm_by_dist=False, **kwargs):

        # the pareto front if necessary to calculate the indicator
        pf = at_least_2d_array(pf, extend_as="row")
        ideal, nadir = derive_ideal_and_nadir_from_pf(pf, ideal=ideal, nadir=nadir)

        super().__init__(zero_to_one=zero_to_one, ideal=ideal, nadir=nadir, **kwargs)
        self.dist_func = dist_func
        self.axis = axis
        self.norm_by_dist = norm_by_dist
        self.pf = self.normalization.forward(pf)

    def _do(self, F):

        # a factor to normalize the distances by (1.0 disables that by default)
        norm = 1.0

        # if zero_to_one is disabled this can be used to normalize the distance calculation itself
        if self.norm_by_dist:
            assert self.ideal is not None and self.nadir is not None, "If norm_by_dist is enabled ideal and nadir must be set!"
            norm = self.nadir - self.ideal

        D = vectorized_cdist(self.pf, F, func_dist=self.dist_func, norm=norm)
        return np.mean(np.min(D, axis=self.axis))



from abc import abstractmethod


class Termination:

    def __init__(self) -> None:
        super().__init__()

        # the algorithm can be forced to terminate by setting this attribute to true
        self.force_termination = False

        # the value indicating how much perc has been made
        self.perc = 0.0

    def update(self, algorithm):
        """
        Provide the termination criterion a current status of the algorithm to update the perc.

        Parameters
        ----------
        algorithm : object
            The algorithm object which is used to determine whether a run has terminated.
        """

        if self.force_termination:
            progress = 1.0
        else:
            progress = self._update(algorithm)
            assert progress >= 0.0

        self.perc = progress
        return self.perc

    def has_terminated(self):
        return self.perc >= 1.0

    def do_continue(self):
        return not self.has_terminated()

    def terminate(self):
        self.force_termination = True

    @abstractmethod
    def _update(self, algorithm):
        pass


class NoTermination(Termination):

    def _update(self, algorithm):
        return 0.0


class MultipleCriteria(Termination):

    def __init__(self, *args) -> None:
        super().__init__()
        self.criteria = args


class TerminateIfAny(MultipleCriteria):

    def _update(self, algorithm):
        return max([termination.update(algorithm) for termination in self.criteria])


class TerminateIfAll(MultipleCriteria):

    def _update(self, algorithm):
        return min([termination.update(algorithm) for termination in self.criteria])



from pymoo.core.algorithm import Algorithm
from pymoo.core.duplicate import DefaultDuplicateElimination, NoDuplicateElimination
from pymoo.core.initialization import Initialization
from pymoo.core.mating import Mating
from pymoo.core.population import Population
from pymoo.core.repair import NoRepair


class GeneticAlgorithm(Algorithm):

    def __init__(self,
                 pop_size=None,
                 sampling=None,
                 selection=None,
                 crossover=None,
                 mutation=None,
                 survival=None,
                 n_offsprings=None,
                 eliminate_duplicates=DefaultDuplicateElimination(),
                 repair=None,
                 mating=None,
                 advance_after_initial_infill=False,
                 **kwargs
                 ):

        super().__init__(**kwargs)

        # the population size used
        self.pop_size = pop_size

        # whether the algorithm should be advanced after initialization of not
        self.advance_after_initial_infill = advance_after_initial_infill

        # the survival for the genetic algorithm
        self.survival = survival

        # number of offsprings to generate through recombination
        self.n_offsprings = n_offsprings

        # if the number of offspring is not set - equal to population size
        if self.n_offsprings is None:
            self.n_offsprings = pop_size

        # set the duplicate detection class - a boolean value chooses the default duplicate detection
        if isinstance(eliminate_duplicates, bool):
            if eliminate_duplicates:
                self.eliminate_duplicates = DefaultDuplicateElimination()
            else:
                self.eliminate_duplicates = NoDuplicateElimination()
        else:
            self.eliminate_duplicates = eliminate_duplicates

        # simply set the no repair object if it is None
        self.repair = repair if repair is not None else NoRepair()

        self.initialization = Initialization(sampling,
                                             repair=self.repair,
                                             eliminate_duplicates=self.eliminate_duplicates)

        if mating is None:
            mating = Mating(selection,
                            crossover,
                            mutation,
                            repair=self.repair,
                            eliminate_duplicates=self.eliminate_duplicates,
                            n_max_iterations=100)
        self.mating = mating

        # other run specific data updated whenever solve is called - to share them in all algorithms
        self.n_gen = None
        self.pop = None
        self.off = None

    def _initialize_infill(self):
        pop = self.initialization.do(self.problem, self.pop_size, algorithm=self)
        return pop

    def _initialize_advance(self, infills=None, **kwargs):
        if self.advance_after_initial_infill:
            self.pop = self.survival.do(self.problem, infills, n_survive=len(infills), algorithm=self, **kwargs)

    def _infill(self):

        # do the mating using the current population
        off = self.mating.do(self.problem, self.pop, self.n_offsprings, algorithm=self)

        # if the mating could not generate any new offspring (duplicate elimination might make that happen)
        if len(off) == 0:
            self.termination.force_termination = True
            return

        # if not the desired number of offspring could be created
        elif len(off) < self.n_offsprings:
            if self.verbose:
                print("WARNING: Mating could not produce the required number of (unique) offsprings!")

        return off

    def _advance(self, infills=None, **kwargs):

        # the current population
        pop = self.pop

        # merge the offsprings with the current population
        if infills is not None:
            pop = Population.merge(self.pop, infills)

        # execute the survival to find the fittest solutions
        self.pop = self.survival.do(self.problem, pop, n_survive=self.pop_size, algorithm=self, **kwargs)



def Cache(func):
    """

    This is a function decorator for class attributes. It just remembers the result of the FIRST function call
    and returns this from there on. Other cashes like LRU are difficult to use because the input can be unhashable
    or bigger numpy arrays. Thus the user has to choose how to use this cache.
    """

    func_name = func.__name__

    def wrapper(self, *args, use_cache=True, set_cache=True, **kwargs):

        if "cache" not in self.__dict__:
            self.__dict__["cache"] = {}

        cache = self.__dict__["cache"]

        if use_cache and func_name in cache:
            return cache[func_name]
        else:

            obj = func(self, *args, **kwargs)

            if set_cache:
                cache[func_name] = obj

            return obj

    return wrapper



# distutils: language = c++
# cython: language_level=2, boundscheck=False, wraparound=False, cdivision=True


import numpy as np
from libcpp.vector cimport vector



def calc_perpendicular_distance(double[:,:] P, double[:,:] L):
    return np.array(c_calc_perpendicular_distance(P, L))


cdef extern from "math.h":
    double sqrt(double m)
    double pow(double base, double exponent)


cdef double c_norm(double[:] v):
    cdef:
        double val
        int i
    val = 0
    for i in range(v.shape[0]):
        val += pow(v[i], 2)
    val = sqrt(val)
    return val


cdef double[:,:] c_calc_perpendicular_distance(double[:,:] P, double[:,:] L):
    cdef :
        int s_L, s_P, n_dim, i, j, k
        double[:,:] M
        vector[double] N
        double norm, dot, perp_dist, norm_scalar_proj

    s_L = L.shape[0]
    s_P = P.shape[0]
    n_dim = L.shape[1]

    M = np.zeros((s_P, s_L), dtype=np.float64)

    for i in range(s_L):

        norm = c_norm(L[i, :])

        N = vector[double](n_dim)
        for k in range(n_dim):
            N[k] = L[i, k] / norm

        for j in range(s_P):

            dot = 0
            for k in range(n_dim):
                dot += L[i, k] * P[j, k]
            norm_scalar_proj = dot / norm

            perp_dist = 0
            for k in range(n_dim):
                perp_dist += pow(norm_scalar_proj * N[k] - P[j, k], 2)
            perp_dist = sqrt(perp_dist)

            M[j, i] = perp_dist

    return M





import numpy as np

from pymoo.core.meta import Meta
from pymoo.core.problem import Problem, ElementwiseEvaluationFunction


class ElementwiseEvaluationFunctionWithGradient(ElementwiseEvaluationFunction):

    def __call__(self, x):
        f = super().__call__

        from pymoo.gradient import TOOLBOX

        if TOOLBOX == "jax.numpy":
            from pymoo.gradient.grad_jax import jax_elementwise_value_and_grad
            out, grad = jax_elementwise_value_and_grad(f, x)

        elif TOOLBOX == "autograd.numpy":
            from pymoo.gradient.grad_autograd import autograd_elementwise_value_and_grad
            out, grad = autograd_elementwise_value_and_grad(f, x)

        for k, v in grad.items():
            out["d" + k] = np.array(v)

        return out


class ElementwiseAutomaticDifferentiation(Meta, Problem):

    def __init__(self, problem, copy=True):
        if not problem.elementwise:
            raise Exception("Elementwise automatic differentiation can only be applied to elementwise problems.")

        super().__init__(problem, copy)
        self.elementwise_func = ElementwiseEvaluationFunctionWithGradient


class AutomaticDifferentiation(Meta, Problem):

    def do(self, x, return_values_of, *args, **kwargs):
        from pymoo.gradient import TOOLBOX

        vals_not_grad = [v for v in return_values_of if not v.startswith("d")]
        f = lambda xp: self.__object__.do(xp, vals_not_grad, *args, **kwargs)

        if TOOLBOX == "jax.numpy":
            from pymoo.gradient.grad_jax import jax_vectorized_value_and_grad
            out, grad = jax_vectorized_value_and_grad(f, x)

        elif TOOLBOX == "autograd.numpy":
            from pymoo.gradient.grad_autograd import autograd_vectorized_value_and_grad
            out, grad = autograd_vectorized_value_and_grad(f, x)

        for k, v in grad.items():
            out["d" + k] = v

        return out



from pymoo.problems.multi.bnh import *
from pymoo.problems.multi.carside import *
from pymoo.problems.multi.ctp import *
from pymoo.problems.multi.dascmop import *
from pymoo.problems.multi.kursawe import *
from pymoo.problems.multi.modact import *
from pymoo.problems.multi.mw import *
from pymoo.problems.multi.osy import *
from pymoo.problems.multi.srn import *
from pymoo.problems.multi.tnk import *
from pymoo.problems.multi.truss2d import *
from pymoo.problems.multi.welded_beam import *
from pymoo.problems.multi.zdt import *




import numpy as np

from pymoo.algorithms.soo.nonconvex.es import ES
from pymoo.core.population import Population, calc_cv
from pymoo.core.survival import Survival
from pymoo.docs import parse_doc_string
from pymoo.util.function_loader import load_function


class StochasticRankingSurvival(Survival):

    def __init__(self, PR):
        super().__init__(filter_infeasible=False)
        self.PR = PR

    def _do(self, problem, pop, *args, n_survive=None, tcv=None, **kwargs):
        assert problem.n_obj == 1, "This stochastic ranking implementation only works for single-objective problems."

        F, G = pop.get("F", "G")
        f = F[:, 0]

        if not problem.has_constraints():
            I = f.argsort()

        else:
            phi = calc_cv(pop)
            J = np.arange(len(phi))
            I = load_function("stochastic_ranking")(f, phi, self.PR, J)

        return pop[I][:n_survive]


class SRES(ES):

    def __init__(self, PF=0.45, **kwargs):
        """
        Stochastic Ranking Evolutionary Strategy (SRES)

        Parameters
        ----------
        PF: float
            The stochastic ranking weight for choosing a random decision while doing the modified bubble sort.
        """
        super().__init__(survival=StochasticRankingSurvival(PF), **kwargs)
        self.PF = PF

    def _advance(self, infills=None, **kwargs):

        # if not all solutions suggested by infill() are evaluated we create a more semi (mu+lambda) algorithm
        if len(infills) < self.pop_size:
            infills = Population.merge(infills, self.pop)

        self.pop = self.survival.do(self.problem, infills, n_survive=self.pop_size)


parse_doc_string(SRES.__init__)



import numpy as np


class Variable(object):

    def __init__(self, value=None, active=True, flag="default") -> None:
        super().__init__()
        self.value = value
        self.flag = flag
        self.active = active

    def sample(self, n=None):
        if n is None:
            return self._sample(1)[0]
        else:
            return self._sample(n)

    def _sample(self, n):
        pass

    def set(self, value):
        self.value = value

    def get(self, **kwargs):
        return self.value


class BoundedVariable(Variable):

    def __init__(self, value=None, bounds=(None, None), strict=None, **kwargs) -> None:
        super().__init__(value=value, **kwargs)
        self.bounds = bounds

        if strict is None:
            strict = bounds
        self.strict = strict


class Real(BoundedVariable):
    vtype = float

    def _sample(self, n):
        low, high = self.bounds
        return np.random.uniform(low=low, high=high, size=n)


class Integer(BoundedVariable):
    vtype = int

    def _sample(self, n):
        low, high = self.bounds
        return np.random.randint(low, high=high+1, size=n)


class Binary(Variable):
    vtype = bool

    def _sample(self, n):
        return np.random.random(size=n) < 0.5


class Choice(Variable):
    vtype = object

    def __init__(self, value=None, options=None, all=None, **kwargs) -> None:
        super().__init__(value=value, **kwargs)
        self.options = options

        if all is None:
            all = options
        self.all = all

    def _sample(self, n):
        return np.random.choice(self.options, size=n)


def get(*args, size=None, **kwargs):
    if len(args) == 0:
        return

    ret = []
    for arg in args:
        v = arg.get(**kwargs) if isinstance(arg, Variable) else arg

        if size is not None:

            if isinstance(v, np.ndarray):
                v = np.reshape(v, size)
            else:
                v = np.full(size, v)

        ret.append(v)

    if len(ret) == 1:
        return ret[0]
    else:
        return tuple(ret)



import matplotlib.pyplot as plt
import numpy as np

from pymoo.visualization.video.callback_video import AnimationCallback


class OneVariableOneObjectiveVisualization(AnimationCallback):

    def __init__(self,
                 n_samples_for_surface=10000,
                 **kwargs):
        super().__init__(**kwargs)
        self.last_pop = None
        self.n_samples_for_surface = n_samples_for_surface

    def do(self, problem, algorithm):

        # check whether the visualization can be done or not - throw exception or simply do nothing
        if problem.n_var != 1 or problem.n_obj != 1:
            raise Exception("This visualization can only be used for problems with one variable and one objective!")

        # draw the problem surface
        xl, xu = problem.bounds()
        _X = np.linspace(xl, xu, self.n_samples_for_surface)
        _F = problem.evaluate(_X)
        plt.plot(_X, _F, label="True", color="black", alpha=0.6)
        plt.ylim(xl[0], xu[0])
        plt.ylim(_F.min(), _F.max())

        pop = algorithm.pop

        X, F, CV = pop.get("X", "F", "CV")
        plt.scatter(X[:, 0], F[:, 0], color="blue", marker="o", s=70)

        is_new = np.full(len(pop), True)
        if self.last_pop is not None:
            for k, ind in enumerate(pop):
                if ind in self.last_pop:
                    is_new[k] = False

        # plot the new population
        if is_new.sum() > 0:
            X, F, CV = pop[is_new].get("X", "F", "CV")
            plt.scatter(X[:, 0], F[:, 0], color="red", marker="*", s=70)

        if hasattr(algorithm, "off") and algorithm.off is not None:
            X, F, CV = algorithm.off.get("X", "F", "CV")
            plt.scatter(X[:, 0], F[:, 0], color="purple", marker="*", s=40)

        plt.title(f"Generation: {algorithm.n_gen}")
        plt.legend()

        # store the current population as the last
        self.last_pop = set(pop)






class Callback:

    def __init__(self) -> None:
        super().__init__()
        self.data = {}
        self.is_initialized = False

    def initialize(self, algorithm):
        pass

    def notify(self, algorithm):
        pass

    def update(self, algorithm):
        return self._update(algorithm)

    def _update(self, algorithm):
        pass

    def __call__(self, algorithm):

        if not self.is_initialized:
            self.initialize(algorithm)
            self.is_initialized = True

        self.notify(algorithm)
        self.update(algorithm)


class CallbackCollection(Callback):

    def __init__(self, *args) -> None:
        super().__init__()
        self.callbacks = args

    def update(self, algorithm):
        [callback.update(algorithm) for callback in self.callbacks]




import numpy as np

from pymoo.core.decomposition import Decomposition


class Tchebicheff(Decomposition):

    def _do(self, F, weights, **kwargs):
        v = np.abs(F - self.utopian_point) * weights
        tchebi = v.max(axis=1)
        return tchebi



import math

import numpy as np

from pymoo.core.selection import Selection
from pymoo.util.misc import random_permuations


class TournamentSelection(Selection):
    """
      The Tournament selection is used to simulated a tournament between individuals. The pressure balances
      greedy the genetic algorithm will be.
    """

    def __init__(self, func_comp=None, pressure=2, **kwargs):
        """

        Parameters
        ----------
        func_comp: func
            The function to compare two individuals. It has the shape: comp(pop, indices) and returns the winner.
            If the function is None it is assumed the population is sorted by a criterium and only indices are compared.

        pressure: int
            The selection pressure to bie applied. Default it is a binary tournament.
        """

        super().__init__(**kwargs)

        # selection pressure to be applied
        self.pressure = pressure

        self.func_comp = func_comp
        if self.func_comp is None:
            raise Exception("Please provide the comparing function for the tournament selection!")

    def _do(self, _, pop, n_select, n_parents=1, **kwargs):
        # number of random individuals needed
        n_random = n_select * n_parents * self.pressure

        # number of permutations needed
        n_perms = math.ceil(n_random / len(pop))

        # get random permutations and reshape them
        P = random_permuations(n_perms, len(pop))[:n_random]
        P = np.reshape(P, (n_select * n_parents, self.pressure))

        # compare using tournament function
        S = self.func_comp(pop, P, **kwargs)

        return np.reshape(S, (n_select, n_parents))


def compare(a, a_val, b, b_val, method, return_random_if_equal=False):
    if method == 'larger_is_better':
        if a_val > b_val:
            return a
        elif a_val < b_val:
            return b
        else:
            if return_random_if_equal:
                return np.random.choice([a, b])
            else:
                return None
    elif method == 'smaller_is_better':
        if a_val < b_val:
            return a
        elif a_val > b_val:
            return b
        else:
            if return_random_if_equal:
                return np.random.choice([a, b])
            else:
                return None
    else:
        raise Exception("Unknown method.")



import numpy as np

from pymoo.core.problem import Problem
from pymoo.util.remote import Remote

DIFFICULTIES = [
    (0.25, 0., 0.), (0., 0.25, 0.), (0., 0., 0.25), (0.25, 0.25, 0.25),
    (0.5, 0., 0.), (0., 0.5, 0.), (0., 0., 0.5), (0.5, 0.5, 0.5),
    (0.75, 0., 0.), (0., 0.75, 0.), (0., 0., 0.75), (0.75, 0.75, 0.75),
    (0., 1.0, 0.), (0.5, 1.0, 0.), (0., 1.0, 0.5), (0.5, 1.0, 0.5)
]


class DASCMOP(Problem):
    def __init__(self, n_obj, n_ieq_constr, difficulty, **kwargs):
        super().__init__(n_var=30,
                         n_obj=n_obj,
                         n_ieq_constr=n_ieq_constr,
                         vtype=float, xl=0., xu=1., **kwargs)

        if isinstance(difficulty, int):
            self.difficulty = difficulty
            if not (1 <= difficulty <= len(DIFFICULTIES)):
                raise Exception(f"Difficulty must be 1 <= difficulty <= {len(DIFFICULTIES)}, but it is {difficulty}!")
            vals = DIFFICULTIES[difficulty - 1]
        else:
            self.difficulty = -1
            vals = difficulty

        self.eta, self.zeta, self.gamma = vals

    def g1(self, X):
        contrib = (X[:, self.n_obj - 1:] - np.sin(0.5 * np.pi * X[:, 0:1])) ** 2
        return contrib.sum(axis=1)[:, None]

    def g2(self, X):
        z = X[:, self.n_obj - 1:] - 0.5
        contrib = z ** 2 - np.cos(20 * np.pi * z)
        return (self.n_var - self.n_obj + 1) + contrib.sum(axis=1)[:, None]

    def g3(self, X):
        j = np.arange(self.n_obj - 1, self.n_var) + 1
        contrib = (X[:, self.n_obj - 1:] - np.cos(0.25 * j / self.n_var * np.pi * (X[:, 0:1] + X[:, 1:2]))) ** 2
        return contrib.sum(axis=1)[:, None]

    def _calc_pareto_front(self, *args, **kwargs):
        return Remote.get_instance().load("pymoo", "pf", "DASCMOP", f"{str(self.__class__.__name__).lower()}_{self.difficulty}.pf")


class DASCMOP1(DASCMOP):
    def __init__(self, difficulty, **kwargs):
        super().__init__(2, 11, difficulty)

    def constraints(self, X, f0, f1, g):
        a = 20.
        b = 2. * self.eta - 1.
        d = 0.5 if self.zeta != 0 else 0.
        if self.zeta > 0:
            e = d - np.log(self.zeta)
        else:
            e = 1e30
        r = 0.5 * self.gamma

        p_k = np.array([[0., 1.0, 0., 1.0, 2.0, 0., 1.0, 2.0, 3.0]])
        q_k = np.array([[1.5, 0.5, 2.5, 1.5, 0.5, 3.5, 2.5, 1.5, 0.5]])

        a_k2 = 0.3
        b_k2 = 1.2
        theta_k = -0.25 * np.pi

        c = np.zeros((X.shape[0], 2 + p_k.shape[1]))

        c[:, 0] = np.sin(a * np.pi * X[:, 0]) - b
        if self.zeta == 1.:
            c[:, 1:2] = 1e-4 - np.abs(e - g)
        else:
            c[:, 1:2] = (e - g) * (g - d)

        c[:, 2:] = (((f0 - p_k) * np.cos(theta_k) - (f1 - q_k) * np.sin(theta_k)) ** 2 / a_k2
                    + ((f0 - p_k) * np.sin(theta_k) + (f1 - q_k) * np.cos(theta_k)) ** 2 / b_k2
                    - r)

        return -1 * c

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g1(X)

        f0 = X[:, 0:1] + g
        f1 = 1.0 - X[:, 0:1] ** 2 + g

        out["F"] = np.column_stack([f0, f1])
        out["G"] = self.constraints(X, f0, f1, g)


class DASCMOP2(DASCMOP1):
    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g1(X)

        f0 = X[:, 0:1] + g
        f1 = 1.0 - np.sqrt(X[:, 0:1]) + g

        out["F"] = np.column_stack([f0, f1])
        out["G"] = self.constraints(X, f0, f1, g)


class DASCMOP3(DASCMOP1):
    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g1(X)

        f0 = X[:, 0:1] + g
        f1 = 1.0 - np.sqrt(X[:, 0:1]) + 0.5 * np.abs(np.sin(5 * np.pi * X[:, 0:1])) + g

        out["F"] = np.column_stack([f0, f1])
        out["G"] = self.constraints(X, f0, f1, g)


class DASCMOP4(DASCMOP1):
    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g2(X)

        f0 = X[:, 0:1] + g
        f1 = 1.0 - X[:, 0:1] ** 2 + g

        out["F"] = np.column_stack([f0, f1])
        out["G"] = self.constraints(X, f0, f1, g)


class DASCMOP5(DASCMOP1):

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g2(X)

        f0 = X[:, 0:1] + g
        f1 = 1.0 - np.sqrt(X[:, 0:1]) + g

        out["F"] = np.column_stack([f0, f1])
        out["G"] = self.constraints(X, f0, f1, g)


class DASCMOP6(DASCMOP1):
    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g2(X)

        f0 = X[:, 0:1] + g
        f1 = 1.0 - np.sqrt(X[:, 0:1]) + 0.5 * np.abs(np.sin(5 * np.pi * X[:, 0:1])) + g

        out["F"] = np.column_stack([f0, f1])
        out["G"] = self.constraints(X, f0, f1, g)


class DASCMOP7(DASCMOP):

    def __init__(self, difficulty_factors, **kwargs):
        super().__init__(3, 7, difficulty_factors)

    def constraints(self, X, f0, f1, f2, g):
        a = 20.
        b = 2. * self.eta - 1.
        d = 0.5 if self.zeta != 0 else 0
        if self.zeta > 0:
            e = d - np.log(self.zeta)
        else:
            e = 1e30
        r = 0.5 * self.gamma

        x_k = np.array([[1.0, 0., 0., 1.0 / np.sqrt(3.0)]])
        y_k = np.array([[0., 1.0, 0., 1.0 / np.sqrt(3.0)]])
        z_k = np.array([[0., 0., 1.0, 1.0 / np.sqrt(3.0)]])

        c = np.zeros((X.shape[0], 3 + x_k.shape[1]))

        c[:, 0] = np.sin(a * np.pi * X[:, 0]) - b
        c[:, 1] = np.cos(a * np.pi * X[:, 1]) - b
        if self.zeta == 1:
            c[:, 2:3] = 1e-4 - np.abs(e - g)
        else:
            c[:, 2:3] = (e - g) * (g - d)

        c[:, 3:] = (f0 - x_k) ** 2 + (f1 - y_k) ** 2 + (f2 - z_k) ** 2 - r ** 2
        return -1 * c

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g2(X)

        f0 = X[:, 0:1] * X[:, 1:2] + g
        f1 = X[:, 1:2] * (1.0 - X[:, 0:1]) + g
        f2 = 1 - X[:, 1:2] + g

        out["F"] = np.column_stack([f0, f1, f2])
        out["G"] = self.constraints(X, f0, f1, f2, g)


class DASCMOP8(DASCMOP7):

    def objectives(self, X, g):
        f0 = np.cos(0.5 * np.pi * X[:, 0:1]) * np.cos(0.5 * np.pi * X[:, 1:2]) + g
        f1 = np.cos(0.5 * np.pi * X[:, 0:1]) * np.sin(0.5 * np.pi * X[:, 1:2]) + g
        f2 = np.sin(0.5 * np.pi * X[:, 0:1]) + g
        return np.column_stack([f0, f1, f2])

    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g2(X)
        F = self.objectives(X, g)
        out["F"] = F
        out["G"] = self.constraints(X, F[:, 0:1], F[:, 1:2], F[:, 2:3], g)


class DASCMOP9(DASCMOP8):
    def _evaluate(self, X, out, *args, **kwargs):
        g = self.g3(X)
        F = self.objectives(X, g)
        out["F"] = F
        out["G"] = self.constraints(X, F[:, 0:1], F[:, 1:2], F[:, 2:3], g)



import numpy as np

from pymoo.core.plot import Plot
from pymoo.docs import parse_doc_string
from pymoo.util.misc import set_if_none


def plot_1d(sc):
    sc.init_figure()
    labels = sc.get_labels()
    ax = sc.ax

    for k, (F, kwargs) in enumerate(sc.to_plot):
        func = getattr(ax, kwargs.pop("mode"))
        func(F, np.zeros_like(F), **kwargs)
        ax.set_xlabel(labels[0])


def plot_2d(sc):
    sc.init_figure()
    labels = sc.get_labels()
    ax = sc.ax

    for k, (F, kwargs) in enumerate(sc.to_plot):
        func = getattr(ax, kwargs.pop("mode"))
        func(F[:, 0], F[:, 1], **kwargs)
        ax.set_xlabel(labels[0])
        ax.set_ylabel(labels[1])

    return sc


def plot_3d(sc, angle):
    sc.init_figure(plot_3D=True)
    labels = sc.get_labels()
    ax = sc.ax

    for k, (F, kwargs) in enumerate(sc.to_plot):

        # here alo `plot_trisurf` is allowed
        func = getattr(ax, kwargs.pop("mode"))
        func(F[:, 0], F[:, 1], F[:, 2], **kwargs)

        ax.xaxis.pane.fill = False
        ax.yaxis.pane.fill = False
        ax.zaxis.pane.fill = False

        ax.set_xlabel(labels[0])
        ax.set_ylabel(labels[1])
        ax.set_zlabel(labels[2])

        if sc.angle is not None:
            ax.view_init(*angle)


def plot_pairwise(sc):
    sc.init_figure(n_rows=sc.n_dim, n_cols=sc.n_dim)
    labels = sc.get_labels()

    for k, (F, kwargs) in enumerate(sc.to_plot):

        assert F.shape[1] >= 2, "A pairwise sc plot needs at least two dimensions."
        mode = kwargs.pop("mode")

        for i in range(sc.n_dim):
            for j in range(sc.n_dim):

                ax = sc.ax[i, j]
                func = getattr(ax, mode)

                if i != j:
                    func(F[:, i], F[:, j], **kwargs)
                    ax.set_xlabel(labels[i])
                    ax.set_ylabel(labels[j])
                else:
                    func(0, 0, s=1, color="white")
                    ax.set_xticks([])
                    ax.set_yticks([])
                    ax.text(0, 0, labels[i], ha='center', va='center', fontsize=20)


class Scatter(Plot):

    def __init__(self,
                 plot_3d=True,
                 angle=(45, 45),
                 **kwargs):
        """

        Scatter Plot

        Parameters
        ----------------

        axis_style : {axis_style}
        labels : {labels}

        Other Parameters
        ----------------

        figsize : {figsize}
        title : {title}
        legend : {legend}
        tight_layout : {tight_layout}

        """

        super().__init__(**kwargs)
        self.angle = angle
        self.plot_3d = plot_3d

    def _do(self):

        # set some default values
        to_plot = []
        for k, (F, v) in enumerate(self.to_plot):
            v = dict(v)
            set_if_none(v, "color", self.colors[k % len(self.colors)])
            set_if_none(v, "alpha", 1.0)

            # this is added to have compatibility to an old version
            # should be removed when the documentation is updated
            if "plot_type" in v:
                name = v.pop("plot_type")

                if name == "line":
                    name = "plot"
                elif name == "surface":
                    name = "plot_trisurf"

                v["mode"] = name
            set_if_none(v, "mode", "scatter")

            to_plot.append([F, v])

        self.to_plot = to_plot

        if self.n_dim == 1:
            plot_1d(self)
        elif self.n_dim == 2:
            plot_2d(self)
        elif self.n_dim == 3 and self.plot_3d:
            plot_3d(self, self.angle)
        else:
            plot_pairwise(self)

        return self


parse_doc_string(Scatter.__init__)



import pymoo.gradient.toolbox as anp

from pymoo.problems.many.dtlz import DTLZ1, DTLZ3


def constraint_dc1(X, a=5, b=0.95):
    G = b - anp.cos(a * anp.pi * X[:, 0])
    return G


def constraints_dc2(gx, a=3, b=0.9):
    G = anp.column_stack([
        b - anp.cos(gx / 100 * anp.pi * a),
        b - anp.exp (-gx / 100)
    ])
    return G


def constraints_dc3(X, gx, a=5, b=0.5):
    Ggx = b - anp.cos(a * anp.pi * gx)
    Gx = b - anp.cos(a * anp.pi * X)
    return anp.column_stack([Ggx, Gx])


class DC1DTLZ1(DTLZ1):
    def __init__(self, n_var=12, n_obj=3, **kwargs):
        super().__init__(n_var, n_obj, **kwargs)
        self.n_ieq_constr = 1

    def _evaluate(self, x, out, *args, **kwargs):
        super()._evaluate(x, out, *args, **kwargs)
        out["G"] = constraint_dc1(x)


class DC1DTLZ3(DTLZ3):
    def __init__(self, n_var=12, n_obj=3, **kwargs):
        super().__init__(n_var, n_obj, **kwargs)
        self.n_ieq_constr = 1

    def _evaluate(self, x, out, *args, **kwargs):
        super()._evaluate(x, out, *args, **kwargs)
        out["G"] = constraint_dc1(x)


class DC2DTLZ1(DTLZ1):
    def __init__(self, n_var=12, n_obj=3, **kwargs):
        super().__init__(n_var, n_obj, **kwargs)
        self.n_ieq_constr = 2

    def _evaluate(self, x, out, *args, **kwargs):
        X_, X_M = x[:, :self.n_obj - 1], x[:, self.n_obj - 1:]
        g = self.g1(X_M)
        out["F"] = self.obj_func(X_, g)
        out["G"] = constraints_dc2(g)


class DC2DTLZ3(DTLZ3):
    def __init__(self, n_var=12, n_obj=3, **kwargs):
        super().__init__(n_var, n_obj, **kwargs)
        self.n_ieq_constr = 2

    def _evaluate(self, x, out, *args, **kwargs):
        X_, X_M = x[:, :self.n_obj - 1], x[:, self.n_obj - 1:]
        g = self.g1(X_M)
        out["F"] = self.obj_func(X_, g, alpha=1)
        out["G"] = constraints_dc2(g)


class DC3DTLZ1(DTLZ1):
    def __init__(self, n_var=12, n_obj=3, **kwargs):
        super().__init__(n_var, n_obj, n_ieq_constr=n_obj, **kwargs)

    def _evaluate(self, x, out, *args, **kwargs):
        X_, X_M = x[:, :self.n_obj - 1], x[:, self.n_obj - 1:]
        g = self.g1(X_M)
        out["F"] = self.obj_func(X_, g)
        out["G"] = constraints_dc3(X_, g)


class DC3DTLZ3(DTLZ3):
    def __init__(self, n_var=12, n_obj=3, **kwargs):
        super().__init__(n_var, n_obj, n_ieq_constr=n_obj, **kwargs)

    def _evaluate(self, x, out, *args, **kwargs):
        X_, X_M = x[:, :self.n_obj - 1], x[:, self.n_obj - 1:]
        g = self.g1(X_M)
        out["F"] = self.obj_func(X_, g, alpha=1)
        out["G"] = constraints_dc3(X_, g)



import numpy as np
import pymoo.gradient.toolbox as anp


from pymoo.util.ref_dirs.construction import ConstructionBasedReferenceDirectionFactory
from pymoo.util.ref_dirs.misc import project_onto_sum_equals_zero_plane, project_onto_unit_simplex_recursive
from pymoo.util.ref_dirs.optimizer import Adam
from pymoo.util.ref_dirs.reduction import ReductionBasedReferenceDirectionFactory
from pymoo.util.reference_direction import ReferenceDirectionFactory, scale_reference_directions


class RieszEnergyReferenceDirectionFactory(ReferenceDirectionFactory):

    def __init__(self,
                 n_dim,
                 n_points,
                 ref_points=None,
                 return_as_tuple=False,
                 n_max_iter=1000,
                 n_until_optimizer_reset=30,
                 sampling="reduction",
                 norm_gradients=True,
                 verify_gradient=False,
                 freeze_edges=False,
                 precision=1e-5,
                 restarts=True,
                 X=None,
                 d=None,
                 callback=None,
                 **kwargs):

        super().__init__(n_dim, **kwargs)

        self.n_points = n_points
        self.n_max_iter = n_max_iter
        self.n_max_not_improved = n_until_optimizer_reset
        self.return_as_tuple = return_as_tuple
        self.sampling = sampling
        self.X = X
        self.ref_points = ref_points
        self.precision = precision
        self.verify_gradient = verify_gradient
        self.norm_gradients = norm_gradients
        self.freeze_edges = freeze_edges
        self.d = d
        self.callback = callback
        self.restarts = restarts

        # experiments have shown that dimensions squared is good value to choose here
        if self.d is None:
            self.d = n_dim * 2

    def _step(self, optimizer, X, freeze=None):
        free = np.logical_not(freeze)

        obj, grad, mutual_dist = calc_potential_energy_with_grad(X, self.d, return_mutual_dist=True)
        # obj, grad = value_and_grad(calc_potential_energy)(X, self.d)

        if self.verify_gradient:
            from autograd import value_and_grad
            obj, grad = calc_potential_energy_with_grad(X, self.d)
            _obj, _grad = value_and_grad(calc_potential_energy)(X, self.d)
            if np.abs(grad - _grad).mean() > 1e-5:
                print("GRADIENT IMPLEMENTATION IS INCORRECT!")

        # set the gradients for frozen points to zero - make them not to move
        if freeze is not None:
            grad[freeze] = 0

        # project the gradient to have a sum of zero - guarantees to stay on the simplex
        proj_grad = project_onto_sum_equals_zero_plane(grad)

        # normalize the gradients by the largest gradient norm
        if self.norm_gradients:
            norm = np.linalg.norm(proj_grad, axis=1)
            proj_grad = (proj_grad / max(norm.max(), 1e-24))

        # apply a step of gradient descent by subtracting the projected gradient with a learning rate
        X = optimizer.next(X, proj_grad)

        # project the out of bounds points back onto the unit simplex
        X[free] = project_onto_unit_simplex_recursive(X[free])

        # because of floating point issues make sure it is on the unit simplex
        X /= X.sum(axis=1)[:, None]

        return X, obj

    def _solve(self, X, F=None, freeze_edges=True):
        n_points = len(X)
        ret, obj = X, np.inf
        n_not_improved = 0

        # get the edge mask
        if freeze_edges:
            freeze = np.any(X < 1e-16, axis=1)
        else:
            freeze = np.full(len(X), False)

        # if additional points to be frozen are provided - add them to the X and mark them as frozen
        if F is not None:
            X = np.row_stack([X, F])
            freeze = np.concatenate([freeze, np.full(len(F), True)])

        # if all points are frozen - simply return it
        if np.all(freeze):
            return X

        # initialize the optimizer for the run
        self.optimizer = Adam(alpha=0.005)

        if self.callback is not None:
            self.callback(self, X)

        # for each iteration of gradient descent
        for i in range(self.n_max_iter):

            # execute one optimization step
            _X, _obj = self._step(self.optimizer, X, freeze=freeze)

            # if it is the current best solution -> store it
            if _obj < obj:
                ret, obj, n_not_improved = _X, _obj, 0
            else:
                n_not_improved += 1

            # evaluate how much the points have moved
            delta = np.sqrt((_X[:n_points] - X[:n_points]) ** 2).mean(axis=1).mean()

            if self.verbose:
                print(i, "objective", _obj, "delta", delta)

            # if there was only a little delta during the last iteration -> terminate
            if delta < self.precision or np.isnan(_obj):
                break

            # reset the optimizer if the objective value has not improved x iterations
            if self.restarts and n_not_improved > self.n_max_not_improved:
                self.optimizer = Adam(alpha=self.optimizer.alpha / 2)
                _X = ret
                n_not_improved = 0

            # otherwise use the new points for the next iteration
            X = _X

            if self.callback is not None:
                self.callback(self, X)

        return ret[:n_points]

    def _do(self):
        X = self.X

        # if no initial points are provided by the user
        if X is None:
            if self.sampling == "reduction":
                X = ReductionBasedReferenceDirectionFactory(self.n_dim,
                                                            self.n_points,
                                                            kmeans=True,
                                                            lexsort=False) \
                    .do()

            elif self.sampling == "construction":
                X = ConstructionBasedReferenceDirectionFactory(self.n_dim,
                                                               self.n_points) \
                    .do()
            else:
                raise Exception("Unknown sampling method. Either reduction or construction.")

        X = self._solve(X, freeze_edges=self.freeze_edges)

        if self.ref_points is not None:
            X, R = self.calc_ref_points(X, self.ref_points)

            if self.return_as_tuple:
                return X, R
            else:
                return np.row_stack([X, R])

        return X

    def calc_ref_points(self, X, ref_points):
        n_points = len(X)

        # the center needed for translations
        centroid = np.full((1, self.n_dim), 1 / self.n_dim)

        R = []

        # for each reference point provided by the user
        for entry in ref_points:
            ref_point, n_points_of_ref = entry.get("coordinates"), entry.get("n_points")
            scale, volume = entry.get("scale"), entry.get("volume")

            if scale is None:
                if volume is None:
                    raise Exception("Either define scale or volume!")
                else:
                    scale = volume ** (self.n_dim - 1)

            # retrieve all points to consider
            _X = np.row_stack([X] + R)

            # translate X to make the simplex to fill the unit
            v = centroid - ref_point
            X_t = _X + v
            X_t = scale_reference_directions(X_t, 1 / scale)

            # find the indices of points which are used as edges
            I = np.where(np.any(X_t < 1e-5, axis=1))[0]

            # create new points in the simplex where reference directions are supposed to be optimized
            _n_points = n_points_of_ref + (n_points - len(I))
            _R = ReductionBasedReferenceDirectionFactory(self.n_dim, n_points=_n_points, kmeans=True,
                                                         lexsort=False).do()

            # detect the edges and just optimize them and shrink later
            outer = np.any(_R == 0, axis=1)
            inner = ~outer

            # rescale the reference directions to be not too close to existing points
            _R = scale_reference_directions(_R, 0.9)

            # optimize just the edges
            _R[outer] = self._solve(_R[outer], F=np.row_stack([X_t[I], _R[inner]]), freeze_edges=False)

            # no set the reference point and freeze it
            # closest_to_centroid = cdist(centroid, _R).argmin()
            # outer[closest_to_centroid] = True
            # inner[closest_to_centroid] = False
            # _R[closest_to_centroid] = centroid

            # now when translated the reference point becomes the centroid - fix that point to be included
            _R[inner] = self._solve(_R[inner], F=np.row_stack([X_t[I], _R[outer]]), freeze_edges=False)

            # rescale and translate them back
            _R_t = scale_reference_directions(_R, scale)
            _R_t = _R_t - v

            # if any point is out of bounds - volume was too large
            if not np.all(_R_t >= 0):

                # get the corner points if not transformed
                V = (np.eye(self.n_dim) - centroid)

                # get the corner points of the ref dir simplex
                P = ref_point + scale * V
                E = centroid + scale * V

                # project because at least is out of points
                P_proj = project_onto_unit_simplex_recursive(np.copy(P))

                for i in range(len(P)):

                    if not np.all(P[i] == P_proj[i]):
                        _R_t = scale_reference_directions(_R, scale)
                        _R_t = _R_t - (E[i] - P_proj[i])

                        if np.all(_R_t >= 0):
                            break

            n_points += n_points_of_ref
            R.append(_R_t)

            # filter out points from to be removed from the original array
            X = X[[i for i in I if i < len(X)]]

        R = np.row_stack(R)

        return X, R


# ---------------------------------------------------------------------------------------------------------
# Energy Functions
# ---------------------------------------------------------------------------------------------------------


def squared_dist(A, B):
    return ((A[:, None] - B[None, :]) ** 2).sum(axis=2)


def calc_potential_energy(A, d):
    i, j = anp.triu_indices(len(A), 1)
    D = anp.sqrt(squared_dist(A, A)[i, j])
    energy = anp.log((1 / D ** d).mean())
    return energy


def calc_potential_energy_with_grad(x, d, return_mutual_dist=False):
    diff = (x[:, None] - x[None, :])
    # calculate the squared euclidean from each point to another
    dist = np.sqrt((diff ** 2).sum(axis=2))

    # make sure the distance to itself does not count
    np.fill_diagonal(dist, np.inf)

    # epsilon which is the smallest distance possible to avoid an overflow during gradient calculation
    eps = 10 ** (-320 / (d + 2))
    b = dist < eps
    dist[b] = eps

    # select only upper triangular matrix to have each mutual distance once
    mutual_dist = dist[np.triu_indices(len(x), 1)]

    # calculate the energy by summing up the squared distances
    energy = (1 / mutual_dist ** d).sum()
    log_energy = - np.log(len(mutual_dist)) + np.log(energy)

    # calculate the gradient
    grad = (-d * diff) / (dist ** (d + 2))[..., None]
    grad = np.sum(grad, axis=1)
    grad /= energy

    ret = [log_energy, grad]
    if return_mutual_dist:
        ret.append(mutual_dist)

    return tuple(ret)



import numpy as np
import numpy as np

from pymoo.core.problem import Problem


class Knapsack(Problem):
    def __init__(self,
                 n_items,  # number of items that can be picked up
                 W,  # weights for each item
                 P,  # profit of each item
                 C,  # maximum capacity
                 ):
        super().__init__(n_var=n_items, n_obj=1, n_ieq_constr=1, xl=0, xu=1, vtype=bool)

        self.W = W
        self.P = P
        self.C = C

    def _evaluate(self, x, out, *args, **kwargs):
        out["F"] = -np.sum(self.P * x, axis=1)
        out["G"] = (np.sum(self.W * x, axis=1) - self.C)


class MultiObjectiveKnapsack(Knapsack):
    def __init__(self, *args):
        super().__init__(*args)

    def _evaluate(self, x, out, *args, **kwargs):
        f1 = - np.sum(self.P * x, axis=1)
        f2 = np.sum(x, axis=1)

        out["F"] = np.column_stack([f1, f2])
        out["G"] = (np.sum(self.W * x, axis=1) - self.C)


def create_random_knapsack_problem(n_items, seed=1, variant="single"):
    np.random.seed(seed)
    P = np.random.randint(1, 100, size=n_items)
    W = np.random.randint(1, 100, size=n_items)
    C = int(np.sum(W) / 10)

    if variant == "single":
        problem = Knapsack(n_items, W, P, C)
    else:
        problem = MultiObjectiveKnapsack(n_items, W, P, C)

    return problem



from pymoo.algorithms.base.bracket import BracketSearch
from pymoo.core.individual import Individual
from pymoo.core.population import Population


class GoldenSectionSearch(BracketSearch):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.left, self.right = None, None
        self.R = (5 ** 0.5 - 1) / 2

    def _initialize_infill(self):
        super()._initialize_infill()
        a, b = self.a, self.b

        # the golden ratio (precomputed constant)
        R = self.R

        # create the left and right in the interval itself
        c, d = Individual(X=b.X - R * (b.X - a.X)), Individual(X=a.X + R * (b.X - a.X))

        # create a population with all four individuals
        pop = Population.create(a, c, d, b)

        self.pop, self.infills = pop, pop

        return pop

    def _advance(self, **kwargs):

        # all the elements in the interval
        a, c, d, b = self.pop

        # the golden ratio (precomputed constant)
        R = self.R

        # if the left solution is better than the right
        if c.F[0] < d.F[0]:

            # make the right to be the new right bound and the left becomes the right
            a, b = a, d
            d = c

            # create a new left individual and evaluate
            c = Individual(X=b.X - R * (b.X - a.X))
            self.evaluator.eval(self.problem, c, algorithm=self)
            self.infills = c

        # if the right solution is better than the left
        else:

            # make the left to be the new left bound and the right becomes the left
            a, b = c, b
            c = d

            # create a new right individual and evaluate
            d = Individual(X=a.X + R * (b.X - a.X))
            self.evaluator.eval(self.problem, d, algorithm=self)
            self.infills = d

        # update the population with all the four individuals
        self.pop = Population.create(a, c, d, b)





# distutils: language = c++
# cython: language_level=2, boundscheck=False, wraparound=False, cdivision=True


import numpy as np
from libcpp.vector cimport vector


# -----------------------------------------------------------
# INTERFACE
# -----------------------------------------------------------

def calc_perpendicular_distance(double[:,:] P, double[:,:] L):
    return np.array(c_calc_perpendicular_distance(P, L))


def pbi(double[:,:] F, double[:,:] weights, double[:] ideal_point, double theta, double eps=1e-10):
    return np.array(c_pbi(F, weights, ideal_point, theta, eps), dtype=np.float64)


def calc_distance_to_weights(F, weights, utopian_point=None):

    if utopian_point is None:
        utopian_point = np.zeros(F.shape[1])

    norm = np.linalg.norm(weights, axis=1)

    d1, d2 = np.zeros(F.shape[0]), np.zeros(F.shape[0])

    F = F - utopian_point
    c_d1(d1, F, weights, norm)
    c_d2(d2, F, weights, d1, norm)

    return d1, d2


# -----------------------------------------------------------
# IMPLEMENTATION
# -----------------------------------------------------------

cdef extern from "math.h":
    double sqrt(double m)
    double pow(double base, double exponent)




cdef double c_norm(double[:] v):
    cdef:
        double val
        int i

    val = 0
    for i in range(v.shape[0]):
        val += pow(v[i], 2)
    val = sqrt(val)
    return val


cdef double[:,:] c_calc_perpendicular_distance(double[:,:] P, double[:,:] L):
    cdef :
        int s_L, s_P, n_dim, i, j, k
        double[:,:] M
        vector[double] N
        double norm, dot, perp_dist, norm_scalar_proj

    s_L = L.shape[0]
    s_P = P.shape[0]
    n_dim = L.shape[1]

    M = np.zeros((s_P, s_L), dtype=np.float64)

    for i in range(s_L):

        norm = c_norm(L[i, :])

        N = vector[double](n_dim)
        for k in range(n_dim):
            N[k] = L[i, k] / norm

        for j in range(s_P):

            dot = 0
            for k in range(n_dim):
                dot += L[i, k] * P[j, k]
            norm_scalar_proj = dot / norm

            perp_dist = 0
            for k in range(n_dim):
                perp_dist += pow(norm_scalar_proj * N[k] - P[j, k], 2)
            perp_dist = sqrt(perp_dist)

            M[j, i] = perp_dist

    return M



cdef vector[double] c_pbi(double[:,:] F, double[:,:] weights, double[:] ideal_point, double theta, double eps):
    cdef:
        double d1, d2, f_max, norm
        int i, j, n_dim
        vector[double] pbi

    n_points = F.shape[0]
    n_obj = F.shape[1]
    pbi = vector[double](n_points)

    for i in range(n_points):

        norm = c_norm(weights[i,:])

        d1 = 0
        for j in range(n_obj):
            d1 += (F[i,j] - ideal_point[j] + eps) * weights[i,j]
        d1 = d1 / norm

        d2 = 0
        for j in range(n_obj):
            d2 += pow(F[i,j] - ideal_point[j] + eps - (d1 * weights[i,j] / norm), 2.0)
        d2 = sqrt(d2)

        pbi[i] = d1 + theta * d2

    return pbi




cdef void c_d1(double[:] d1, double[:,:] F, double[:,:] weights, double[:] norm):
    cdef:
        double val
        int i, j

    n_points = F.shape[0]
    n_obj = F.shape[1]

    for i in range(n_points):

        val = 0
        for j in range(n_obj):
            val += F[i,j] * weights[i,j]
        d1[i] = val / norm[i]




cdef void c_d2(double[:] d2, double[:,:] F, double[:,:] weights, double[:] d1, double[:] norm):
    cdef:
        double val
        int i, j

    n_points = F.shape[0]
    n_obj = F.shape[1]

    for i in range(n_points):

        val = 0
        for j in range(n_obj):
            val += pow(F[i,j] - (d1[i] * weights[i,j] / norm[i]), 2.0)
        d2[i] = sqrt(val)







from pymoo.core.algorithm import Algorithm
from pymoo.core.individual import Individual


class BracketSearch(Algorithm):

    def __init__(self, a=None, b=None, **kwargs):
        super().__init__(**kwargs)
        self.a, self.b = a, b

    def _setup(self, problem, a=None, b=None, **kwargs):
        msg = "Only problems with one variable, one objective and no constraints can be solved!"
        assert problem.n_var == 1 and not problem.has_constraints() and problem.n_obj == 1, msg
        self.a, self.b = a, b

    def _initialize_infill(self):

        # the boundaries of the problem for initialization
        xl, xu = self.problem.bounds()

        # take care of the lower bound - make sure it is an individual and make sure it is evaluated
        if self.a is None:
            assert xl is not None, "Either provide a left bound or set the xl attribute in the problem."
            self.a = Individual(X=xl)

        if self.a.F is None:
            self.evaluator.eval(self.problem, self.a, algorithm=self)

        # take care of the upper bound - make sure it is an individual and make sure it is evaluated
        if self.b is None:
            assert xl is not None, "Either provide a right bound or set the xu attribute in the problem."
            self.b = Individual(X=xu)

        if self.b.F is None:
            self.evaluator.eval(self.problem, self.b, algorithm=self)

        assert self.a.X[0] <= self.b.X[0], "The left bound must be smaller than the left bound!"




import weakref

import numpy as np



class Tree:
    '''
    Implementation of Nary-tree.
    The source code is modified based on https://github.com/lianemeth/forest/blob/master/forest/NaryTree.py

    Parameters
    ----------
    key: object
        key of the node
    num_branch: int
        how many branches in each node
    children: Iterable[Tree]
        reference of the children
    parent: Tree
        reference of the parent node
    Returns
    -------
        an N-ary tree.
    '''

    def __init__(self, key, num_branch, children=None, parent=None):
        self.key = key
        self.children = children or [None for _ in range(num_branch)]

        self._parent = weakref.ref(parent) if parent else None

    @property
    def parent(self):
        if self._parent:
            return self._parent()

    def __getstate__(self):
        self._parent = None

    def __setstate__(self, state):
        self.__dict__ = state
        for child in self.children:
            child._parent = weakref.ref(self)

    def traversal(self, visit=None, *args, **kwargs):
        if visit is not None:
            visit(self, *args, **kwargs)
        l = [self]
        for child in self.children:
            if child is not None:
                l += child.traversal(visit, *args, **kwargs)
        return l


def tree_based_non_dominated_sort(F):
    """
    Tree-based efficient non-dominated sorting (T-ENS).
    This algorithm is very efficient in many-objective optimization problems (MaOPs).
    Parameters
    ----------
    F: np.array
        objective values for each individual.
    Returns
    -------
        indices of the individuals in each front.
    References
    ----------
    X. Zhang, Y. Tian, R. Cheng, and Y. Jin,
    A decision variable clustering based evolutionary algorithm for large-scale many-objective optimization,
    IEEE Transactions on Evolutionary Computation, 2018, 22(1): 97-112.
    """

    N, M = F.shape
    # sort the rows in F
    indices = np.lexsort(F.T[::-1])
    F = F[indices]

    obj_seq = np.argsort(F[:, :0:-1], axis=1) + 1

    k = 0

    forest = []

    left = np.full(N, True)
    while np.any(left):
        forest.append(None)
        for p, flag in enumerate(left):
            if flag:
                update_tree(F, p, forest, k, left, obj_seq)
        k += 1

    # convert forest to fronts
    fronts = [[] for _ in range(k)]
    for k, tree in enumerate(forest):
        fronts[k].extend([indices[node.key] for node in tree.traversal()])
    return fronts


def update_tree(F, p, forest, k, left, obj_seq):
    _, M = F.shape
    if forest[k] is None:
        forest[k] = Tree(key=p, num_branch=M - 1)
        left[p] = False
    elif check_tree(F, p, forest[k], obj_seq, True):
        left[p] = False


def check_tree(F, p, tree, obj_seq, add_pos):
    if tree is None:
        return True

    N, M = F.shape

    # find the minimal index m satisfying that p[obj_seq[tree.root][m]] < tree.root[obj_seq[tree.root][m]]
    m = 0
    while m < M - 1 and F[p, obj_seq[tree.key, m]] >= F[tree.key, obj_seq[tree.key, m]]:
        m += 1

    # if m not found
    if m == M - 1:
        # p is dominated by the solution at the root
        return False
    else:
        for i in range(m + 1):
            # p is dominated by a solution in the branch of the tree
            if not check_tree(F, p, tree.children[i], obj_seq, i == m and add_pos):
                return False

        if tree.children[m] is None and add_pos:
            # add p to the branch of the tree
            tree.children[m] = Tree(key=p, num_branch=M - 1)
        return True



import numpy as np

from pymoo.core.crossover import Crossover
from pymoo.util.misc import crossover_mask


class UniformCrossover(Crossover):

    def __init__(self, **kwargs):
        super().__init__(2, 2, **kwargs)

    def _do(self, _, X, **kwargs):
        _, n_matings, n_var = X.shape
        M = np.random.random((n_matings, n_var)) < 0.5
        _X = crossover_mask(X, M)
        return _X


class UX(UniformCrossover):
    pass



import pymoo.gradient.toolbox as anp
import numpy as np

from pymoo.core.problem import Problem
from pymoo.util.remote import Remote


class TNK(Problem):
    def __init__(self):
        super().__init__(n_var=2, n_obj=2, n_ieq_constr=2, vtype=float)
        self.xl = np.array([0, 1e-30])
        self.xu = np.array([np.pi, np.pi])

    def _evaluate(self, x, out, *args, **kwargs):
        f1 = x[:, 0]
        f2 = x[:, 1]
        g1 = -(anp.square(x[:, 0]) + anp.square(x[:, 1]) - 1.0 - 0.1 * anp.cos(16.0 * anp.arctan(x[:, 0] / x[:, 1])))
        g2 = 2 * (anp.square(x[:, 0] - 0.5) + anp.square(x[:, 1] - 0.5)) - 1

        out["F"] = anp.column_stack([f1, f2])
        out["G"] = anp.column_stack([g1, g2])

    def _calc_pareto_front(self, *args, **kwargs):
        return Remote.get_instance().load("pymoo", "pf", "tnk.pf")



import copy


def minimize(problem, algorithm, termination=None, copy_algorithm=True, copy_termination=True, **kwargs):
    """

    Minimization of function of one or more variables, objectives and constraints.

    This is used as a convenience function to execute several algorithms with default settings which turned
    out to work for a test single. However, evolutionary computations utilizes the idea of customizing a
    meta-algorithm. Customizing the algorithm using the object oriented interface is recommended to improve the
    convergence.

    Parameters
    ----------

    problem : :class:`~pymoo.core.problem.Problem`
        A problem object which is defined using pymoo.

    algorithm : :class:`~pymoo.core.algorithm.Algorithm`
        The algorithm object that should be used for the optimization.

    termination : :class:`~pymoo.core.termination.Termination` or tuple
        The termination criterion that is used to stop the algorithm.

    seed : integer
        The random seed to be used.

    verbose : bool
        Whether output should be printed or not.

    display : :class:`~pymoo.util.display.Display`
        Each algorithm has a default display object for printouts. However, it can be overwritten if desired.

    callback : :class:`~pymoo.core.callback.Callback`
        A callback object which is called each iteration of the algorithm.

    save_history : bool
        Whether the history should be stored or not.

    copy_algorithm : bool
        Whether the algorithm object should be copied before optimization.

    Returns
    -------
    res : :class:`~pymoo.core.result.Result`
        The optimization result represented as an object.

    """

    # create a copy of the algorithm object to ensure no side-effects
    if copy_algorithm:
        algorithm = copy.deepcopy(algorithm)

    # initialize the algorithm object given a problem - if not set already
    if algorithm.problem is None:
        if termination is not None:

            if copy_termination:
                termination = copy.deepcopy(termination)

            kwargs["termination"] = termination

        algorithm.setup(problem, **kwargs)

    # actually execute the algorithm
    res = algorithm.run()

    # store the deep copied algorithm in the result object
    res.algorithm = algorithm

    return res



import numpy as np

from pymoo.algorithms.base.local import LocalSearch
from pymoo.algorithms.moo.nsga2 import RankAndCrowdingSurvival
from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.util.display.single import SingleObjectiveOutput

from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting
from pymoo.util.normalization import normalize, denormalize


def norm_bounds(pop, problem):
    nxl = normalize(pop.get("xl"), problem.xl, problem.xu)
    nxu = normalize(pop.get("xu"), problem.xl, problem.xu)
    return nxl, nxu


def update_bounds(ind, xl, xu, k, delta):
    _xl = np.copy(xl)
    _xl[k] = ind.X[k] - delta
    ind.set("xl", _xl)

    _xu = np.copy(xu)
    _xu[k] = ind.X[k] + delta
    ind.set("xu", _xu)


class DIRECT(LocalSearch):

    def __init__(self,
                 eps=1e-2,
                 penalty=0.1,
                 n_max_candidates=10,
                 n_max_archive=400,
                 archive_reduct=0.66,
                 output=SingleObjectiveOutput(),
                 **kwargs):
        super().__init__(output=output, **kwargs)
        self.eps = eps
        self.penalty = penalty
        self.n_max_candidates = n_max_candidates
        self.n_max_archive = n_max_archive
        self.archive_reduct = archive_reduct

    def _setup(self, problem, **kwargs):

        xl, xu = problem.bounds()
        X = denormalize(0.5 * np.ones(problem.n_var), xl, xu)

        x0 = Individual(X=X)
        x0.set("xl", xl)
        x0.set("xu", xu)
        x0.set("depth", 0)

        self.x0 = x0

    def _initialize_infill(self, **kwargs):
        return Population.create(self.x0)

    def _potential_optimal(self):
        pop = self.pop

        if len(pop) == 1:
            return pop

        # get the intervals of each individual
        _F, _CV, xl, xu = pop.get("F", "CV", "xl", "xu")
        nF = normalize(_F)
        F = nF + self.penalty * _CV

        # get the length of the interval of each solution
        nxl, nxu = norm_bounds(pop, self.problem)
        length = (nxu - nxl) / 2
        val = length.mean(axis=1)

        # (a) non-dominated set with respect to interval
        obj = np.column_stack([-val, F])

        # an unlimited archive size can cause issues - thus truncate if necessary
        if len(pop) > self.n_max_archive:
            # find the rank of each individual
            _, rank = NonDominatedSorting().do(obj, return_rank=True)

            # calculate the number of solutions after truncation and filter the best ones out
            n_truncated = int(self.archive_reduct * self.n_max_archive)
            I = np.argsort(rank)[:n_truncated]

            # also update all the utility variables defined so far to match the truncation
            pop, F, nxl, nxu, length, val, obj = pop[I], F[I], nxl[I], nxu[I], length[I], val[I], obj[I]
            self.pop = pop

        I = NonDominatedSorting().do(obj, only_non_dominated_front=True)
        candidates, F, xl, xu, val = pop[I], F[I], xl[I], xu[I], val[I]

        # if all candidates are expanded in each iteration this can cause issues - here use crowding distance to decide
        if len(candidates) == 1:
            return candidates
        else:
            if len(candidates) > self.n_max_candidates:
                candidates = RankAndCrowdingSurvival().do(self.problem, pop, n_survive=self.n_max_candidates)

            return candidates

    def _infill(self):

        # the offspring population to finally evaluate and attach to the population
        infills = Population()

        # find the potential optimal solution in the current population
        potential_optimal = self._potential_optimal()

        # for each of those solutions execute the division move
        for current in potential_optimal:

            # find the largest dimension the solution has not been evaluated yet
            nxl, nxu = norm_bounds(current, self.problem)
            k = np.argmax(nxu - nxl)

            # the delta value to be used to get left and right - this is one sixth of the range
            xl, xu = current.get("xl"), current.get("xu")

            delta = (xu[k] - xl[k]) / 6

            # create the left individual
            left_x = np.copy(current.X)
            left_x[k] = xl[k] + delta
            left = Individual(X=left_x)

            # create the right individual
            right_x = np.copy(current.X)
            right_x[k] = xu[k] - delta
            right = Individual(X=right_x)

            # update the boundaries for all the points accordingly
            for ind in [current, left, right]:
                update_bounds(ind, xl, xu, k, delta)

            # create the offspring population, evaluate and attach to current population
            _infill = Population.create(left, right)
            _infill.set("depth", current.get("depth") + 1)

            infills = Population.merge(infills, _infill)

        return infills

    def _advance(self, infills=None, **kwargs):
        assert infills is not None, "This algorithms uses the AskAndTell interface thus infills must to be provided."
        self.pop = Population.merge(self.pop, infills)


import math

from pymoo.core.infill import InfillCriterion


class Mating(InfillCriterion):

    def __init__(self,
                 selection,
                 crossover,
                 mutation,
                 **kwargs):

        super().__init__(**kwargs)
        self.selection = selection
        self.crossover = crossover
        self.mutation = mutation

    def _do(self, problem, pop, n_offsprings, parents=None, **kwargs):

        # how many parents need to be select for the mating - depending on number of offsprings remaining
        n_matings = math.ceil(n_offsprings / self.crossover.n_offsprings)

        # if the parents for the mating are not provided directly - usually selection will be used
        if parents is None:

            # select the parents for the mating - just an index array
            parents = self.selection(problem, pop, n_matings, n_parents=self.crossover.n_parents, **kwargs)

        # do the crossover using the parents index and the population - additional data provided if necessary
        off = self.crossover(problem, parents, **kwargs)

        # do the mutation on the offsprings created through crossover
        off = self.mutation(problem, off, **kwargs)

        return off






import numpy as np

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival
from pymoo.core.duplicate import DefaultDuplicateElimination, DuplicateElimination
from pymoo.core.population import Population
from pymoo.core.selection import Selection
from pymoo.core.survival import Survival
from pymoo.docs import parse_doc_string
from pymoo.operators.crossover.binx import BinomialCrossover
from pymoo.operators.mutation.nom import NoMutation
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.selection.rnd import RandomSelection
from pymoo.termination.default import DefaultSingleObjectiveTermination
from pymoo.util.display.single import SingleObjectiveOutput
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting


# =========================================================================================================
# Implementation
# =========================================================================================================


class EliteSurvival(Survival):

    def __init__(self, n_elites, eliminate_duplicates=None):
        super().__init__(False)
        self.n_elites = n_elites
        self.eliminate_duplicates = eliminate_duplicates

    def _do(self, problem, pop, n_survive=None, algorithm=None, **kwargs):

        if isinstance(self.eliminate_duplicates, bool) and self.eliminate_duplicates:
            pop = DefaultDuplicateElimination(func=lambda p: p.get("F")).do(pop)

        elif isinstance(self.eliminate_duplicates, DuplicateElimination):
            _, no_candidates, candidates = DefaultDuplicateElimination(func=lambda pop: pop.get("F")).do(pop,
                                                                                                         return_indices=True)
            _, _, is_duplicate = self.eliminate_duplicates.do(pop[candidates], pop[no_candidates], return_indices=True,
                                                              to_itself=False)
            elim = set(np.array(candidates)[is_duplicate])
            pop = pop[[k for k in range(len(pop)) if k not in elim]]

        if problem.n_obj == 1:
            pop = FitnessSurvival().do(problem, pop, n_survive=len(pop))
            elites = pop[:self.n_elites]
            non_elites = pop[self.n_elites:]
        else:
            I = NonDominatedSorting().do(pop.get("F"), only_non_dominated_front=True)
            elites = pop[I]
            non_elites = pop[[k for k in range(len(pop)) if k not in I]]

        elites.set("type", ["elite"] * len(elites))
        non_elites.set("type", ["non_elite"] * len(non_elites))

        return pop


class EliteBiasedSelection(Selection):

    def _do(self, problem, pop, n_select, n_parents, **kwargs):
        _type = pop.get("type")
        elites = np.where(_type == "elite")[0]
        non_elites = np.where(_type == "non_elite")[0]

        # if through duplicate elimination no non-elites exist
        if len(non_elites) == 0:
            non_elites = elites

        # do the mating selection - always one elite and one non-elites
        s_elite = np.random.choice(elites, size=n_select)
        s_non_elite = np.random.choice(non_elites, size=n_select)

        return np.column_stack([s_elite, s_non_elite])


class BRKGA(GeneticAlgorithm):

    def __init__(self,
                 n_elites=200,
                 n_offsprings=700,
                 n_mutants=100,
                 bias=0.7,
                 sampling=FloatRandomSampling(),
                 survival=None,
                 output=SingleObjectiveOutput(),
                 eliminate_duplicates=False,
                 **kwargs
                 ):
        """


        Parameters
        ----------

        n_elites : int
            Number of elite individuals

        n_offsprings : int
            Number of offsprings to be generated through mating of an elite and a non-elite individual

        n_mutants : int
            Number of mutations to be introduced each generation

        bias : float
            Bias of an offspring inheriting the allele of its elite parent

        eliminate_duplicates : bool or class
            The duplicate elimination is more important if a decoding is used. The duplicate check has to be
            performed on the decoded variable and not on the real values. Therefore, we recommend passing
            a DuplicateElimination object.
            If eliminate_duplicates is simply set to `True`, then duplicates are filtered out whenever the
            objective values are equal.

        """

        if survival is None:
            survival = EliteSurvival(n_elites, eliminate_duplicates=eliminate_duplicates)

        super().__init__(pop_size=n_elites + n_offsprings + n_mutants,
                         n_offsprings=n_offsprings,
                         sampling=sampling,
                         selection=EliteBiasedSelection(),
                         crossover=BinomialCrossover(bias, prob=1.0),
                         mutation=NoMutation(),
                         survival=survival,
                         output=output,
                         eliminate_duplicates=True,
                         advance_after_initial_infill=True,
                         **kwargs)

        self.n_elites = n_elites
        self.n_mutants = n_mutants
        self.bias = bias
        self.termination = DefaultSingleObjectiveTermination()

    def _infill(self):
        pop = self.pop

        # actually do the mating given the elite selection and biased crossover
        off = self.mating.do(self.problem, pop, n_offsprings=self.n_offsprings, algorithm=self)

        # create the mutants randomly to fill the population with
        mutants = FloatRandomSampling().do(self.problem, self.n_mutants, algorithm=self)

        # evaluate all the new solutions
        return Population.merge(off, mutants)

    def _advance(self, infills=None, **kwargs):
        pop = self.pop

        # get all the elites from the current population
        elites = np.where(pop.get("type") == "elite")[0]

        # finally merge everything together and sort by fitness
        pop = Population.merge(pop[elites], infills)

        # the do survival selection - set the elites for the next round
        self.pop = self.survival.do(self.problem, pop, n_survive=len(pop), algorithm=self)


parse_doc_string(BRKGA.__init__)



import matplotlib.pyplot as plt
import numpy as np

from pymoo.core.callback import Callback
from pymoo.indicators.hv import Hypervolume
from pymoo.indicators.igd import IGD
from pymoo.termination.ftol import calc_delta_norm
from pymoo.util.normalization import normalize
from pymoo.util.sliding_window import SlidingWindow
from pymoo.visualization.video.callback_video import AnimationCallback


class RunningMetric(Callback):

    def __init__(self,
                 period=None,
                 indicator="igd") -> None:

        super().__init__()
        self.indicator = indicator

        self.delta_ideal = None
        self.delta_nadir = None
        self.delta_f = None

        self.history = SlidingWindow(period)

    def update(self, algorithm):
        history = self.history

        F = algorithm.opt.get("F")
        c_F, c_ideal, c_nadir = F, F.min(axis=0), F.max(axis=0)

        # append the current optimum to the history
        history.append(dict(F=F, ideal=c_ideal, nadir=c_nadir))

        # the current norm that should be used for normalization
        norm = c_nadir - c_ideal
        norm[norm < 1e-32] = 1.0

        # normalize the current objective space values
        c_N = normalize(c_F, c_ideal, c_nadir)

        # normalize all previous generations with respect to current ideal and nadir
        N = [normalize(e["F"], c_ideal, c_nadir) for e in history]

        # calculate the delta difference for each previous ideal and nadir point to current
        delta_ideal = [calc_delta_norm(history[k]["ideal"], history[k-1]["ideal"], norm) for k in range(1, len(history))] + [0.0]
        delta_nadir = [calc_delta_norm(history[k]["nadir"], history[k-1]["nadir"], norm) for k in range(1, len(history))] + [0.0]

        # now calculate the indicator from each previous one to the current
        if self.indicator == "igd":
            delta_f = [IGD(c_N).do(N[k]) for k in range(len(N))]
        elif self.indicator == "hv":
            hv = Hypervolume(ref_point=np.ones(c_F.shape[1]))
            delta_f = [hv.do(N[k]) for k in range(len(N))]
        else:
            raise Exception("Unknown indicator.")

        self.delta_ideal, self.delta_nadir, self.delta_f = delta_ideal, delta_nadir, delta_f


class RunningMetricAnimation(AnimationCallback):

    def __init__(self,
                 delta_gen,
                 n_plots=4,
                 key_press=True,
                 **kwargs) -> None:

        super().__init__(**kwargs)
        self.running = RunningMetric()
        self.delta_gen = delta_gen
        self.key_press = key_press
        self.data = SlidingWindow(n_plots)

    def draw(self, data, ax):

        for tau, x, f, v in data[:-1]:
            ax.plot(x, f, label="t=%s" % tau, alpha=0.6, linewidth=3)

        tau, x, f, v = data[-1]
        ax.plot(x, f, label="t=%s (*)" % tau, alpha=0.9, linewidth=3)

        for k in range(len(v)):
            if v[k]:
                ax.plot([k + 1, k + 1], [0, f[k]], color="black", linewidth=0.5, alpha=0.5)
                ax.plot([k + 1], [f[k]], "o", color="black", alpha=0.5, markersize=2)

        ax.set_yscale("symlog")
        ax.legend()

        ax.set_xlabel("Generation")
        ax.set_ylabel("$\Delta \, f$", rotation=0)

    def do(self, _, algorithm, force_plot=False, **kwargs):
        running = self.running

        # update the running metric to have the most recent information
        running.update(algorithm)

        tau = algorithm.n_gen

        if (tau > 0 and tau % self.delta_gen == 0) or force_plot:

            f = running.delta_f
            x = np.arange(len(f)) + 1
            v = [max(ideal, nadir) > 0.005 for ideal, nadir in zip(running.delta_ideal, running.delta_nadir)]
            self.data.append((tau, x, f, v))

            fig, ax = plt.subplots()
            self.draw(self.data, ax)

            if self.key_press:
                def press(event):
                    if event.key == 'q':
                        algorithm.termination.force_termination = True

                fig.canvas.mpl_connect('key_press_event', press)

                plt.draw()
                plt.waitforbuttonpress()
                plt.close('all')




import numpy as np
import pymoo.gradient.toolbox as anp

from pymoo.core.problem import Problem


class Ackley(Problem):

    def __init__(self, n_var=2, a=20, b=1/5, c=2 * np.pi):
        super().__init__(n_var=n_var, n_obj=1, xl=-32.768, xu=+32.768, vtype=float)
        self.a = a
        self.b = b
        self.c = c

    def _evaluate(self, x, out, *args, **kwargs):
        part1 = -1. * self.a * anp.exp(-1. * self.b * anp.sqrt((1. / self.n_var) * anp.sum(x * x, axis=1)))
        part2 = -1. * anp.exp((1. / self.n_var) * anp.sum(anp.cos(self.c * x), axis=1))
        out["F"] = part1 + part2 + self.a + anp.exp(1)

    def _calc_pareto_front(self):
        return 0

    def _calc_pareto_set(self):
        return np.full(self.n_var, 0)



from copy import deepcopy

import numpy as np

from pymoo.core.operator import Operator
from pymoo.core.variable import Real, get


class Mutation(Operator):

    def __init__(self, prob=1.0, prob_var=None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.prob = Real(prob, bounds=(0.7, 1.0), strict=(0.0, 1.0))
        self.prob_var = Real(prob_var, bounds=(0.0, 0.25), strict=(0.0, 1.0)) if prob_var is not None else None

    def do(self, problem, pop, inplace=True, **kwargs):

        # if not inplace copy the population first
        if not inplace:
            pop = deepcopy(pop)

        n_mut = len(pop)

        # get the variables to be mutated
        X = pop.get("X")

        # retrieve the mutation variables
        Xp = self._do(problem, X, **kwargs)

        # the likelihood for a mutation on the individuals
        prob = get(self.prob, size=n_mut)
        mut = np.random.random(size=n_mut) <= prob

        # store the mutated individual back to the population
        pop[mut].set("X", Xp[mut])

        return pop

    def _do(self, problem, X, **kwargs):
        return X

    def get_prob_var(self, problem, **kwargs):
        prob_var = self.prob_var if self.prob_var is not None else min(0.5, 1 / problem.n_var)
        return get(prob_var, **kwargs)



import numpy as np


class RouletteWheelSelection:

    def __init__(self, val, larger_is_better=True):
        super().__init__()
        if not larger_is_better:
            val = val.max() - val
        _sum = val.sum()
        self.cumulative = np.array([val[:k].sum() / _sum for k in range(1, len(val))])

    def next(self, n=None):
        if n is None:
            X = np.random.random((1, 1))
        else:
            X = np.random.random((n, 1))
            if n > 1:
                X.repeat(n - 1, axis=1)

        M = self.cumulative[None, :].repeat(len(X), axis=0)
        B = X >= M
        ret = B.sum(axis=1)

        if n is None:
            return ret[0]
        return ret



from pymoo.util.display.column import Column
from pymoo.util.display.output import Output, pareto_front_if_possible


class MinimumConstraintViolation(Column):

    def __init__(self, **kwargs) -> None:
        super().__init__("cv_min", **kwargs)

    def update(self, algorithm):
        self.value = algorithm.opt.get("cv").min()


class AverageConstraintViolation(Column):

    def __init__(self, **kwargs) -> None:
        super().__init__("cv_avg", **kwargs)

    def update(self, algorithm):
        self.value = algorithm.pop.get("cv").mean()


class SingleObjectiveOutput(Output):

    def __init__(self):
        super().__init__()
        self.cv_min = MinimumConstraintViolation()
        self.cv_avg = AverageConstraintViolation()

        self.f_min = Column(name="f_min")
        self.f_avg = Column(name="f_avg")
        self.f_gap = Column(name="f_gap")

        self.best = None

    def initialize(self, algorithm):
        problem = algorithm.problem

        if problem.has_constraints():
            self.columns += [self.cv_min, self.cv_avg]

        self.columns += [self.f_avg, self.f_min]

        pf = pareto_front_if_possible(problem)
        if pf:
            self.best = pf.flatten()[0]
            self.columns += [self.f_gap]

    def update(self, algorithm):
        super().update(algorithm)

        f, cv, feas = algorithm.pop.get("f", "cv", "feas")

        if feas.sum() > 0:
            self.f_avg.set(f[feas].mean())
        else:
            self.f_avg.set(None)

        opt = algorithm.opt[0]

        if opt.feas:
            self.f_min.set(opt.f)
            if self.best:
                self.f_gap.set(opt.f - self.best)

        else:
            self.f_min.set(None)
            self.f_gap.set(None)



import numpy as np

from pymoo.indicators.hv import hvc_looped
from pymoo.util.function_loader import load_function


def hv_exact(ref_point, F):
    func = load_function("hv")
    hv = func(ref_point, F)
    return hv


def hvc_exact_loopwise(ref_point, F):
    return hvc_looped(ref_point, F, hv_exact)


class DynamicHypervolume:

    def __init__(self, ref_point, F=None, func_hv=None, func_hvc=None) -> None:
        super().__init__()
        self.ref_point = ref_point

        self.func_hv = func_hv
        self.func_hvc = func_hvc

        self.n_dim = len(ref_point)
        self.F = np.zeros((0, self.n_dim))

        self.hv = 0.0
        self.hvc = np.zeros(0)

        if F is not None:
            self.add(F)

    def add(self, F):
        assert len(F.shape) == 2, "The points to add must be a two-dimensional array."
        assert F.shape[1] == self.n_dim, "The dimensions of the ref_point and points to add must be equal"
        self.F = np.row_stack([self.F, F])
        self.hv, self.hvc = self.calc()
        return self

    def delete(self, k):
        assert k < len(self.F)
        self.F = np.delete(self.F, k, axis=0)
        self.hv, self.hvc = self.calc()
        return self

    def calc(self):
        return self._calc(self.ref_point, self.F)

        # if len(self.F) == 0:
        #     return 0.0, np.zeros(0)
        # else:
        #     return self._calc(self.ref_point, self.F)

    def _calc(self, ref_point, F):
        hv = None
        if self.func_hv is not None:
            hv = self.func_hv(ref_point, F)

        hvc = None
        if self.func_hvc is not None:
            hvc = self.func_hvc(ref_point, F)

        return hv, hvc


class ExactHypervolume(DynamicHypervolume):

    def __init__(self, ref_point, func_hv=hv_exact, func_hvc=hvc_exact_loopwise, **kwargs) -> None:
        super().__init__(ref_point, func_hv=func_hv, func_hvc=func_hvc, **kwargs)
